#!/usr/bin/env python
import csv
import sys
from datetime import datetime, timedelta, timezone

import duo_client

argv_iter = iter(sys.argv[1:])


def get_next_arg(prompt, default=None):
    try:
        return next(argv_iter)
    except StopIteration:
        return input(prompt) or default


today = datetime.now(tz=timezone.utc)
default_mintime = int((today - timedelta(days=180)).timestamp())
default_maxtime = int(today.timestamp() * 1000) - 120

# Configuration and information about objects to create.
admin_api = duo_client.Admin(
    ikey=get_next_arg("Admin API integration key: "),
    skey=get_next_arg("Integration secret key: "),
    host=get_next_arg("API hostname: "),
)
params = {}

mintime = get_next_arg("Mintime: ", default_mintime)
if mintime:
    params["mintime"] = mintime

maxtime = get_next_arg("Maxtime: ", default_maxtime)
if maxtime:
    params["maxtime"] = maxtime

limit = get_next_arg("Limit (1000): ")
if limit:
    params["limit"] = limit

next_offset = get_next_arg("Next_offset: ")
if next_offset:
    params["next_offset"] = next_offset

sort = get_next_arg("Sort (ts:desc): ")
if sort:
    params["sort"] = sort

log_type = get_next_arg("Log Type (telephony_v2): ", "telephony_v2")
print(f"Fetching {log_type} logs...")
reporter = csv.writer(sys.stdout)

print("==============================")
if log_type == "activity":
    params["mintime"] = params["mintime"] * 1000
    activity_logs = admin_api.get_activity_logs(**params)
    print(
        "Next offset from response: ", activity_logs.get("metadata").get("next_offset")
    )
    reporter.writerow(
        ("activity_id", "ts", "action", "actor_name", "target_name", "application")
    )
    for log in activity_logs["items"]:
        activity = log["activity_id"]
        ts = log["ts"]
        action = log["action"]
        actor_name = log.get("actor", {}).get("name", None)
        target_name = log.get("target", {}).get("name", None)
        application = log.get("application", {}).get("name", None)
        reporter.writerow(
            [
                activity,
                ts,
                action,
                actor_name,
                target_name,
                application,
            ]
        )
if log_type == "telephony_v2":
    telephony_logs = admin_api.get_telephony_log(api_version=2, kwargs=params)
    reporter.writerow(("telephony_id", "txid", "credits", "context", "phone", "type"))
    
    for log in telephony_logs["items"]:
        telephony_id = log["telephony_id"]
        txid = log["txid"]
        credits = log["credits"]
        context = log["context"]
        phone = log["phone"]
        type = log["type"]
        reporter.writerow(
            [
                telephony_id,
                txid,
                credits,
                context,
                phone,
                type
            ]
        )
if log_type == "auth":
    auth_logs = admin_api.get_authentication_log(api_version=2, kwargs=params)
    print(
        "Next offset from response: ",
        auth_logs.get("metadata").get("next_offset"),
    )
    reporter.writerow(("admin", "akey", "context", "phone", "provider"))
    for log in auth_logs["authlogs"]:
        admin = log["admin_name"]
        akey = log["akey"]
        context = log["context"]
        phone = log["phone"]
        provider = log["provider"]
        reporter.writerow(
            [
                admin,
                akey,
                context,
                phone,
                provider,
            ]
        )

print("==============================")



#!/usr/bin/env python
import sys
import json
import duo_client


argv_iter = iter(sys.argv[1:])


def get_next_arg(prompt):
    try:
        return next(argv_iter)
    except StopIteration:
        return input(prompt)


admin_api = duo_client.Admin(
    ikey=get_next_arg('Admin API integration key ("DI..."): '),
    skey=get_next_arg("integration secret key: "),
    host=get_next_arg('API hostname ("api-....duosecurity.com"): '),
)


def create_empty_policy(name, print_response=False):
    """
    Create an empty policy with a specified name.
    """

    json_request = {
        "policy_name": name,
    }
    response = admin_api.create_policy_v2(json_request)
    if print_response:
        pretty = json.dumps(response, indent=4, sort_keys=True, default=str)
        print(pretty)
    return response.get("policy_key")


def create_policy_browsers(name, print_response=False):
    """
    Create a policy that blocks internet explorer browsers. Requires
    Access or Beyond editions.
    """

    json_request = {
        "policy_name": name,
        "sections": {
            "browsers": {
                "blocked_browsers_list": [
                    "ie",
                ],
            },
        },
    }
    response = admin_api.create_policy_v2(json_request)
    if print_response:
        pretty = json.dumps(response, indent=4, sort_keys=True, default=str)
        print(pretty)
    return response.get("policy_key")

def copy_policy(name1, name2, copy_from, print_response=False):
    """
    Copy the policy `copy_from` to two new policies.
    """
    response = admin_api.copy_policy_v2(copy_from, [name1, name2])
    if print_response:
        pretty = json.dumps(response, indent=4, sort_keys=True, default=str)
        print(pretty)
    policies = response.get("policies")
    return (policies[0].get("policy_key"), policies[1].get("policy_key"))

def bulk_delete_section(policy_keys, print_response=False):
    """
    Delete the section "browsers" from the provided policies.
    """
    response = admin_api.update_policies_v2("", ["browsers"], policy_keys)
    if print_response:
        pretty = json.dumps(response, indent=4, sort_keys=True, default=str)
        print(pretty)

def update_policy_with_device_health_app(policy_key, print_response=False):
    """
    Update a given policy to include Duo Device Health App policy
    settings. Requires Access or Beyond editions.
    NOTE: this function is deprecated, please use update_policy_with_duo_desktop
    """
    return update_policy_with_duo_desktop(policy_key, print_response)

def update_policy_with_duo_desktop(policy_key, print_response=False):
    """
    Update a given policy to include Duo Desktop policy
    settings. Requires Access or Beyond editions.
    """

    json_request = {
        "sections": {
            "duo_desktop": {
                "enforce_encryption": ["windows"],
                "enforce_firewall": ["windows"],
                "requires_duo_desktop": ["windows"],
                "windows_endpoint_security_list": ["cisco-amp"],
                "windows_remediation_note": "Please install Windows agent",
            },
        },
    }
    response = admin_api.update_policy_v2(policy_key, json_request)
    if print_response:
        pretty = json.dumps(response, indent=4, sort_keys=True, default=str)
        print(pretty)
    return response.get("policy_key")


def get_policy(policy_key):
    """
    Fetch a given policy.
    """

    response = admin_api.get_policy_v2(policy_key)
    pretty = json.dumps(response, indent=4, sort_keys=True, default=str)
    print(pretty)


def iterate_all_policies():
    """
    Loop over each policy.
    """

    print("#####################")
    print("Iterating over all policies...")
    print("#####################")
    iter = sorted(
        admin_api.get_policies_v2_iterator(), key=lambda x: x.get("policy_name")
    )
    for policy in iter:
        print(
            "##################### {} {}".format(
                policy.get("policy_name"), policy.get("policy_key")
            )
        )
        pretty = json.dumps(policy, indent=4, sort_keys=True, default=str)
        print(pretty)


def main():
    # Create two empty policies
    policy_key_a = create_empty_policy("Test New Policy - a")
    policy_key_b = create_empty_policy("Test New Policy - b")

    # Update policy with Duo Desktop settings.
    update_policy_with_duo_desktop(policy_key_b)

    # Create an empty policy and delete it.
    policy_key_c = create_empty_policy("Test New Policy - c")
    admin_api.delete_policy_v2(policy_key_c)

    # Create a policy with browser restriction settings.
    policy_key_d = create_policy_browsers("Test New Policy - d")

    # Copy a policy to 2 new policies.
    policy_key_e, policy_key_f = copy_policy("Test New Policy - e", "Test New Policy - f", policy_key_d)

    # Delete the browser restriction settings from 2 policies.
    bulk_delete_section([policy_key_e, policy_key_f])

    # Fetch the global and other custom policy.
    get_policy("global")
    get_policy(policy_key_b)

    # Loop over each policy.
    iterate_all_policies()


if __name__ == "__main__":
    main()



"""
Script to pull list of all phones and modify the name of each
"""

import sys
import getpass

import duo_client

argv_iter = iter(sys.argv[1:])


def _get_next_arg(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    try:
        return next(argv_iter)
    except StopIteration:
        if secure is True:
            return getpass.getpass(prompt)
        else:
            return input(prompt)


admin_api = duo_client.Admin(
        ikey=_get_next_arg('Admin API integration key ("DI..."): '),
        skey=_get_next_arg('integration secret key: ', secure=True),
        host=_get_next_arg('API hostname ("api-....duosecurity.com"): '),
)

phones = admin_api.get_phones()

for phone in phones:
    print(f"Current phone name for device ID {phone['phone_id']} is {phone['name']}")
    new_phone_name = phone['name'] + '_new'
    print(f"Changing name to {new_phone_name}")
    result = admin_api.update_phone(phone_id=phone['phone_id'], name=new_phone_name)
    if result['name'] == new_phone_name:
        print(f"Device {phone['phone_id']} is now named {new_phone_name}.")
    else:
        print("An error occurred.")




"""
Example of creating an integration in a child account using parent account credentials
"""

import argparse
import duo_client


parser = argparse.ArgumentParser()
duo_arg_group = parser.add_argument_group('Duo Accounts API Credentials')
duo_arg_group.add_argument('--ikey',
                           help='Duo Accounts API IKEY',
                           required=True
                           )
duo_arg_group.add_argument('--skey',
                           help='Duo Accounts API Secret Key',
                           required=True,
                           )
duo_arg_group.add_argument('--host',
                           help='Duo child account API apihost',
                           required=True
                           )
parser.add_argument('--child_account_id',
                    help='The Duo account ID of the child account to query.',
                    required=True
                    )
args = parser.parse_args()

# It is important to note that we are using the IKEY/SKEY combination for an Accounts API integration in the
# parent account along with the api-hostname of a child account to create a new duo_client.Admin instance
account_client = duo_client.Admin(
        ikey=args.ikey,
        skey=args.skey,
        host=args.host,
)

# Once the duo_client.Admin instance is created, the child account_id is assigned. This is necessary to ensure
# queries made with this Admin API instance are directed to the proper child account that matches the api-hostname
# used to create the instance.
account_client.account_id = args.child_account_id


def main():
    """Main program entry point"""

    print(f"Retrieving integrations for child account {args.child_account_id}")
    child_account_integrations = account_client.get_integrations_generator()
    for integration in child_account_integrations:
        print(f"{integration['name']=}")


if __name__ == '__main__':
    main()



from .. import util
import duo_client.admin
from .base import TestAdmin


class TestU2F(TestAdmin):
    def test_get_u2ftokens_with_params(self):
        """ Test to get u2ftokens with params.
        """
        response = list(self.client_list.get_u2ftokens(limit=8))[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/u2ftokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id':[self.client_list.account_id],
                'limit': ['8'],
                'offset': ['0'],
            }
        )

    def test_get_u2ftokens_iterator(self):
        response = self.client_list.get_u2ftokens_iterator()
        response = next(response)
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/u2ftokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0']
            }
        )

    def test_get_u2ftokens_without_params(self):
        """ Test to get u2ftokens without params.
        """
        response = list(self.client_list.get_u2ftokens())[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/u2ftokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
             }
        )

    def test_get_u2ftokens_with_offset(self):
        response = list(self.client_list.get_u2ftokens(limit=2, offset=3))[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/u2ftokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id':[self.client_list.account_id],
                'limit': ['2'],
                'offset': ['3']
            }
        )

    def test_get_u2ftoken_by_id(self):
        """ Test to get u2ftoken by registration id.
        """
        response = self.client.get_u2ftoken_by_id('DU012345678901234567')
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/u2ftokens/DU012345678901234567')
        self.assertEqual(util.params_to_dict(args),
                         {'account_id':[self.client.account_id]})

    def test_delete_u2ftoken(self):
        """ Test to delete u2ftoken by registration id.
        """
        response = self.client.delete_u2ftoken('DU012345678901234567')
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'DELETE')
        self.assertEqual(uri, '/admin/v1/u2ftokens/DU012345678901234567')
        self.assertEqual(util.params_to_dict(args),
                         {'account_id':[self.client.account_id]})



"""
Example of how to extract users and their aliases for a specific group from the Duo Admin API
"""

import sys
import argparse
import dataclasses
from collections import deque
from duo_client import Admin

DUO_MAX_USERS_PER_API_CALL = 100


@dataclasses.dataclass
class DuoUser:
    """
    Duo User object class for storage and retrieval of pertinent information per user
    """
    username: str = None
    user_id: str = None
    group_name: str = None
    aliases: str = None

    def set_group_name(self, group_name):
        self.group_name = group_name

    def set_aliases(self, aliases: list):
        self.aliases = ','.join(aliases)

    def get_user_info(self):
        return_string = (f"'username': {self.username}, 'user_id': {self.user_id}, " +
                         f"'group_name': {self.group_name}, 'aliases': '{self.aliases}'")
        return return_string


parser = argparse.ArgumentParser()
duo_arg_group = parser.add_argument_group('Duo Admin API Credentials')
duo_arg_group.add_argument('--ikey',
                           help='Duo Admin API IKEY',
                           required=True
                           )
duo_arg_group.add_argument('--skey',
                           help='Duo Admin API Secret Key',
                           required=True,
                           )
duo_arg_group.add_argument('--host',
                           help='Duo Admin API apihost',
                           required=True
                           )
parser.add_argument('--group_name',
                    help="Name of group to get users from. Groups are case-sensitive.",
                    required=True
                    )
args = parser.parse_args()

duo_admin_client = Admin(
        ikey=args.ikey,
        skey=args.skey,
        host=args.host
)


def split_list(input_list: list, size: int) -> list:
    """Split a list into chunks based on size"""
    return [input_list[i:i + size] for i in range(0, len(input_list), size)]


def get_duo_group_users(group_name: str) -> list:
    """Get the list of users assigned to the given group name"""
    group_id = None
    try:
        groups = (duo_admin_client.get_groups())
    except Exception as e_str:
        print(f"Exception while retrieving groups: {e_str}")
        sys.exit(1)

    for group in groups:
        if group['name'] == group_name:
            group_id = group['group_id']
            break

    user_list = list(deque(duo_admin_client.get_group_users_iterator(group_id)))
    return split_list(user_list, DUO_MAX_USERS_PER_API_CALL)


def get_duo_user_aliases(user_list: list[list]) -> list:
    """Collect aliases for the users in the given list"""
    all_users = []
    for u_list in user_list:
        users = list(deque(duo_admin_client.get_users_by_ids([uid['user_id'] for uid in u_list])))
        for user in users:
            new_duo_user = DuoUser(user_id=user['user_id'], username=user['username'])
            new_duo_user.set_aliases(list(user['aliases'].values()))
            all_users.append(new_duo_user)
    return all_users


def output_user_aliases(user_list: list[DuoUser], group_name: str) -> None:
    """Output the list of users and their aliases for the requested group"""
    for user in user_list:
        user.set_group_name(group_name)
        print(user.get_user_info())


def main():
    """Main program entry point"""
    if args.group_name is not None:
        duo_group_users = get_duo_group_users(args.group_name)
        if len(duo_group_users) == 0:
            print(f"Unable to find users assigned to group named '{args.group_name}'.")
            sys.exit(1)
        duo_group_user_aliases = get_duo_user_aliases(duo_group_users)
        output_user_aliases(duo_group_user_aliases, args.group_name)


if __name__ == '__main__':
    main()



from .. import util
import duo_client.admin
from .base import TestAdmin


class TestTokens(TestAdmin):
    def test_get_tokens_generator(self):
        """ Test to get tokens generator.
        """
        generator = self.client_list.get_tokens_generator()
        response = next(generator)
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/tokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_tokens(self):
        """ Test to get tokens without pagination params.
        """
        response = self.client_list.get_tokens()
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/tokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_tokens_with_limit(self):
        """ Test to get tokens with pagination params.
        """
        response = self.client_list.get_tokens(limit=20)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/tokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['0'],
            })

    def test_get_tokens_with_limit_offset(self):
        """ Test to get tokens with pagination params.
        """
        response = self.client_list.get_tokens(limit=20, offset=2)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/tokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['2'],
            })

    def test_get_tokens_with_offset(self):
        """ Test to get tokens with pagination params.
        """
        response = self.client_list.get_tokens(offset=9001)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/tokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })



import json
from .. import util
import duo_client.admin
from .base import TestAdmin


class TestUsers(TestAdmin):
    def test_get_users_generator(self):
        """ Test to get users iterator.
        """
        iterator = self.client_list.get_users_iterator()
        response = next(iterator)
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/users')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_users(self):
        """ Test to get users.
        """
        response = self.client_list.get_users()[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/users')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_users_offset(self):
        """ Test to get users with pagination params.
        """
        response = self.client_list.get_users(offset=30)[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/users')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_users_limit(self):
        """ Test to get users with pagination params.
        """
        response = self.client_list.get_users(limit=30)[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/users')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['30'],
                'offset': ['0'],
            })

    def test_get_users_limit_and_offset(self):
        """ Test to get users with pagination params.
        """
        response = self.client_list.get_users(limit=20, offset=30)[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/users')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['30'],
            })

    # GET with params
    def test_get_users_by_name(self):
        response = self.client.get_users_by_name('foo')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users')
        self.assertEqual(
            util.params_to_dict(args),
            {'username':['foo'],
             'account_id':[self.client.account_id]})
        self.assertEqual(response['body'], None)
        response = self.client.get_users_by_name('foo')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users')
        self.assertEqual(
            util.params_to_dict(args),
            {'username':['foo'],
             'account_id':[self.client.account_id]})
        self.assertEqual(response['body'], None)

    # POST with params
    def test_add_user(self):
        # all params given
        response = self.client.add_user(
            'foo', realname='bar', status='active', notes='notes',
            email='foobar@baz.com', firstname='fName', lastname='lName',
            alias1='alias1', alias2='alias2', alias3='alias3', alias4='alias4')
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/admin/v1/users')
        self.assertEqual(
            json.loads(response['body']),
            {
                'realname': 'bar',
                'notes': 'notes',
                'username': 'foo',
                'status': 'active',
                'email': 'foobar@baz.com',
                'firstname': 'fName',
                'lastname': 'lName',
                'account_id': self.client.account_id,
                'alias1': 'alias1',
                'alias2': 'alias2',
                'alias3': 'alias3',
                'alias4': 'alias4',
            })
        # defaults
        response = self.client.add_user('bar')
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/admin/v1/users')
        self.assertEqual(
            json.loads(response['body']),
            {'username':'bar', 'account_id':self.client.account_id})

    def test_update_user(self):
        response = self.client.update_user(
            'DU012345678901234567', username='foo', realname='bar',
            status='active', notes='notes', email='foobar@baz.com',
            firstname='fName', lastname='lName', alias1='alias1',
            alias2='alias2', alias3='alias3', alias4='alias4')
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(
            response['uri'], '/admin/v1/users/DU012345678901234567')
        self.assertEqual(
            json.loads(response['body']),
            {
                'account_id':self.client.account_id,
                'realname': 'bar',
                'notes': 'notes',
                'username': 'foo',
                'status': 'active',
                'email': 'foobar@baz.com',
                'firstname': 'fName',
                'lastname': 'lName',
                'account_id': self.client.account_id,
                'alias1': 'alias1',
                'alias2': 'alias2',
                'alias3': 'alias3',
                'alias4': 'alias4',
            })

    def test_sync_user(self):
        """ Test to synchronize a single user in a directory for a username.
        """
        response = self.client.sync_user('foo', 'test_dir_key')
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'],
                         '/admin/v1/users/directorysync/test_dir_key/syncuser')
        self.assertEqual(
           json.loads(response['body']),
            {'username': 'foo', 'account_id': self.client.account_id})



"""
Low level functions for generating Duo Web API calls and parsing results.
"""
__version__ = '5.4.0'

import base64
import collections
import datetime
import email.utils
import hashlib
import hmac
import http.client
import json
import os
import random
from time import sleep
import socket
import ssl
import sys
import urllib.parse

try:
    # For the optional demonstration CLI program.
    import argparse
except ImportError as e:
    argparse_error = e
    argparse = None

try:
    # Only needed if signing requests with timezones other than UTC.
    import pytz
except ImportError as e:
    pytz = None
    pytz_error = e

from .https_wrapper import CertValidatingHTTPSConnection

DEFAULT_CA_CERTS = os.path.join(os.path.dirname(__file__), 'ca_certs.pem')


def canon_params(params):
    """
    Return a canonical string version of the given request parameters.
    """
    # this is normalized the same as for OAuth 1.0,
    # http://tools.ietf.org/html/rfc5849#section-3.4.1.3.2
    args = []
    for (key, vals) in sorted(
        (urllib.parse.quote(key, '~'), vals) for (key, vals) in list(params.items())):
        for val in sorted(urllib.parse.quote(val, '~') for val in vals):
            args.append('%s=%s' % (key, val))
    return '&'.join(args)


def canon_x_duo_headers(additional_headers):
    """
    Args:
        additional_headers: Dict
    Returns:
        stringified version of all headers that start with 'X-Duo*'. Which is then hashed.
        Note: the keys are also lower-cased for signing.
    """
    if additional_headers is None:
        additional_headers = {}

    # Lower the headers before sorting them
    lowered_headers = {}
    for header_name, header_value in additional_headers.items():
        header_name = header_name.lower() if header_name is not None else None
        lowered_headers[header_name] = header_value

    canon_list = []
    added_headers = []  # store headers we've added, use for duplicate checking (case insensitive)
    for header_name in sorted(lowered_headers.keys()):
        # Extract header value and set key to lower case from now on.
        value = lowered_headers[header_name]

        # Validation gate. We will raise if a problem is found here.
        _validate_additional_header(header_name, value, added_headers)

        # Add to the list of values to canonicalize:
        canon_list.extend([header_name, value])
        added_headers.append(header_name)

    canon = '\x00'.join(canon_list)
    return hashlib.sha512(canon.encode('utf-8')).hexdigest()


def _validate_additional_header(header_name, value, added_headers):
    """
    Args:
        header_name: str
        value: str
        added_headers: list[str] - headers we've already added - check for duplicates (case insensitive)
    Returns: None

    Validates additional headers added to request - headers must comply with the following rules (for V5 sig_version)
    """
    if header_name is None or value is None:
        raise ValueError("Not allowed 'None' as a header name or value")
    if '\x00' in header_name:
        raise ValueError("Not allowed 'Null' character in header name")
    if '\x00' in value:
        raise ValueError("Not allowed 'Null' character in header value")
    if not header_name.lower().startswith('x-duo-'):
        raise ValueError("Additional headers must start with \'X-Duo-\'")
    if header_name.lower() in added_headers:
        raise ValueError("Duplicate header passed, header={}".format(header_name))


def canonicalize(method, host, uri, params, date, sig_version, body=None, additional_headers=None):
    """
    Return a canonical string version of the given request attributes.

    * method: string HTTP method
    * host: string hostname
    * uri: string uri path
    * params: string containing request params
    * date: date string for request
    * sig_version: signature version integer
    * body: request body, must be string for sig_version 4
    """
    if sig_version == 1:
        canon = [
            method.upper(),
            host.lower(),
            uri,
            canon_params(params),
        ]
    elif sig_version == 2:
        canon = [
            date,
            method.upper(),
            host.lower(),
            uri,
            canon_params(params),
        ]
    elif sig_version == 4:
        # sig_version 4 is json only
        canon = [
            date,
            method.upper(),
            host.lower(),
            uri,
            canon_params(params),
            hashlib.sha512(body.encode('utf-8')).hexdigest(),
        ]
    elif sig_version == 5:
        canon = [
            date,
            method.upper(),
            host.lower(),
            uri,
            canon_params(params),
            hashlib.sha512(body.encode('utf-8')).hexdigest(),
            canon_x_duo_headers(additional_headers),  # hashed in canon_x_duo_headers
        ]
    else:
        raise ValueError("Unknown signature version: {}".format(sig_version))
    return '\n'.join(canon)


def sign(ikey, skey, method, host, uri, date, sig_version, params, body=None,
         digestmod=hashlib.sha512, additional_headers=None):
    """
    Return basic authorization header line with a Duo Web API signature.
    """
    canonical = canonicalize(method, host, uri, params, date, sig_version, body=body, additional_headers=additional_headers)
    if isinstance(skey, str):
        skey = skey.encode('utf-8')
    if isinstance(canonical, str):
        canonical = canonical.encode('utf-8')

    sig = hmac.new(skey, canonical, digestmod)
    auth = '%s:%s' % (ikey, sig.hexdigest())

    if isinstance(auth, str):
        auth = auth.encode('utf-8')
    b64 = base64.b64encode(auth)
    if not isinstance(b64, str):
        b64 = b64.decode('utf-8')
    return 'Basic %s' % b64


def normalize_params(params):
    """
    Return copy of params with strings listified
    and unicode strings utf-8 encoded.
    """
    # urllib cannot handle unicode strings properly. quote() excepts,
    # and urlencode() replaces them with '?'.
    def encode(value):
        if isinstance(value, bool):
            if value:
                value = 'true'
            else:
                value = 'false'
        elif isinstance(value, int):
            value = str(value)
        if isinstance(value, str):
            return value.encode("utf-8")
        return value
    def to_list(value):
        if value is None or isinstance(value, str):
            return [value]
        return value
    return dict(
        (encode(key), [encode(v) for v in to_list(value)])
        for (key, value) in list(params.items()))


class Client(object):
    sig_version = 5

    def __init__(self, ikey, skey, host,
                 ca_certs=DEFAULT_CA_CERTS,
                 sig_timezone='UTC',
                 user_agent=('Duo API Python/' + __version__),
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 paging_limit=100,
                 digestmod=hashlib.sha512,
                 sig_version=None,
                 port=None
                 ):
        """
        ca_certs - Path to CA pem file.
        """
        self.ikey = ikey
        self.skey = skey
        self.host = host
        self.port = port
        self.sig_timezone = sig_timezone
        if ca_certs is None:
            ca_certs = DEFAULT_CA_CERTS
        self.ca_certs = ca_certs
        self.user_agent = user_agent
        self.set_proxy(host=None, proxy_type=None)
        self.paging_limit = paging_limit
        self.digestmod = digestmod
        if sig_version is not None:
            self.sig_version = sig_version

        # Constants for handling rate limit backoff and retries
        self._MAX_BACKOFF_WAIT_SECS = 32
        self._INITIAL_BACKOFF_WAIT_SECS = 1
        self._BACKOFF_FACTOR = 2
        self._RATE_LIMITED_RESP_CODE = 429

        # Default timeout is a sentinel object
        if timeout is socket._GLOBAL_DEFAULT_TIMEOUT:
            self.timeout = timeout
        else:
            self.timeout = float(timeout)

        if sig_version == 3:
            raise ValueError('sig_version 3 not supported')

    def set_proxy(self, host, port=None, headers=None,
                  proxy_type='CONNECT'):
        """
        Configure proxy for API calls. Supported proxy_type values:

        'CONNECT' - HTTP proxy with CONNECT.
        None - Disable proxy.
        """
        if proxy_type not in ('CONNECT', None):
            raise NotImplementedError('proxy_type=%s' % (proxy_type,))
        self.proxy_headers = headers
        self.proxy_host = host
        self.proxy_port = port
        self.proxy_type = proxy_type

    def api_call(
        self,
        method,
        path,
        params,
        additional_headers=None,
        sig_version=None,
    ):
        """
        Call a Duo API method. Return a (response, data) tuple.

        * method: HTTP request method. E.g. "GET", "POST", or "DELETE".
        * path: Full path of the API endpoint. E.g. "/auth/v2/ping".
        * params: dict mapping from parameter name to stringified value,
            or a dict to be converted to json.
        * sig_version: signature version integer
        """
        params_go_in_body = method in ('POST', 'PUT', 'PATCH')
        digestmod = self.digestmod
        if additional_headers is None:
            additional_headers = {}
        if sig_version is None:
            sig_version = self.sig_version

        if sig_version in (1, 2):
            params = normalize_params(params)
            # v1 and v2 canonicalization don't distinguish between
            # params and body. There's no separate body input.
            body = None
        elif sig_version in (4, 5):
            digestmod = hashlib.sha512
            if params_go_in_body:
                body = self.canon_json(params)
                params = {}
            else:
                body = ''
                params = normalize_params(params)
        else:
            raise ValueError('unsupported sig_version {}'.format(sig_version))

        if self.sig_timezone == 'UTC':
            now = email.utils.formatdate()
        elif pytz is None:
            raise pytz_error
        else:
            d = datetime.datetime.now(pytz.timezone(self.sig_timezone))
            now = d.strftime("%a, %d %b %Y %H:%M:%S %z")

        auth = sign(self.ikey,
                    self.skey,
                    method,
                    self.host,
                    path,
                    now,
                    sig_version,
                    params,
                    body=body,
                    digestmod=digestmod,
                    additional_headers=additional_headers)
        headers = {
            'Authorization': auth,
            'Date': now,
        }

        if sig_version == 5:
            for k, v in additional_headers.items():
                headers[k] = v

        if self.user_agent:
            headers['User-Agent'] = self.user_agent

        if params_go_in_body:
            if sig_version in (4, 5):
                headers['Content-type'] = 'application/json'
            else:
                headers['Content-type'] = 'application/x-www-form-urlencoded'
                body = urllib.parse.urlencode(params, doseq=True)
            uri = path
        else:
            body = None
            uri = path + '?' + urllib.parse.urlencode(params, doseq=True)

        encoded_headers = {}
        for k, v in headers.items():
            if isinstance(k, str):
                k = k.encode('ascii')
            if isinstance(v, str):
                v = v.encode('ascii')
            encoded_headers[k] = v

        return self._make_request(method, uri, body, encoded_headers)

    def _connect(self):
        # Host and port for the HTTP(S) connection to the API server.
        if self.ca_certs == 'HTTP':
            api_port = 80
        else:
            api_port = 443
        if self.port is not None:
            api_port = self.port

        # Host and port for outer HTTP(S) connection if proxied.
        if self.proxy_type is None:
            host = self.host
            port = api_port
        elif self.proxy_type == 'CONNECT':
            host = self.proxy_host
            port = self.proxy_port
        else:
            raise NotImplementedError('proxy_type=%s' % (self.proxy_type,))

        # Create outer HTTP(S) connection.
        if self.ca_certs == 'HTTP':
            conn = http.client.HTTPConnection(host, port)
        elif self.ca_certs == 'DISABLE':
            kwargs = {}
            if hasattr(ssl, '_create_unverified_context'):
                # httplib.HTTPSConnection validates certificates by
                # default in Python 2.7.9+.
                kwargs['context'] = ssl._create_unverified_context()  # noqa: DUO122, explicitly disabled for testing scenarios
            conn = http.client.HTTPSConnection(host, port, **kwargs)
        else:
            conn = CertValidatingHTTPSConnection(host,
                                                 port,
                                                 ca_certs=self.ca_certs)

        # Override default socket timeout if requested.
        conn.timeout = self.timeout

        # Configure CONNECT proxy tunnel, if any.
        if self.proxy_type == 'CONNECT':
            if hasattr(conn, 'set_tunnel'): # 2.7+
                conn.set_tunnel(self.host,
                                api_port,
                                self.proxy_headers)
            elif hasattr(conn, '_set_tunnel'): # 2.6.3+
                # pylint: disable=E1103
                conn._set_tunnel(self.host,
                                 api_port,
                                 self.proxy_headers)
                # pylint: enable=E1103

        return conn

    def _make_request(self, method, uri, body, headers):
        if self.proxy_type == 'CONNECT':
            # Ensure the request uses the correct protocol and Host.
            if self.ca_certs == 'HTTP':
                api_proto = 'http'
            else:
                api_proto = 'https'
            uri = ''.join((api_proto, '://', self.host, uri))
        conn = self._connect()

        # backoff on rate limited requests and retry. if a request is rate
        # limited after MAX_BACKOFF_WAIT_SECS, return the rate limited response
        wait_secs = self._INITIAL_BACKOFF_WAIT_SECS
        while True:
            response, data = self._attempt_single_request(
                conn, method, uri, body, headers)
            if (response.status != self._RATE_LIMITED_RESP_CODE or
                    wait_secs > self._MAX_BACKOFF_WAIT_SECS):
                break
            random_offset = random.uniform(0.0, 1.0)  # noqa: DUO102, non-cryptographic random use
            sleep(wait_secs + random_offset)
            wait_secs = wait_secs * self._BACKOFF_FACTOR

        self._disconnect(conn)
        return (response, data)

    def _attempt_single_request(self, conn, method, uri, body, headers):
        conn.request(method, uri, body, headers)
        response = conn.getresponse()
        data = response.read()
        return (response, data)

    def _disconnect(self, conn):
        conn.close()

    def normalize_paging_args(self, limit=None, offset=0):
        """
        Converts paging arguments to a format the rest of the client expects.

        :param limit: The number of objects requested of a paginated api
                      endpoint. If it looks falsy, it is is not changed.
                      Default None
        :param offset: The offset to start retrieval. Default 0
        :return: tuple after the form of (limit, offset)
        """

        if limit:
            limit = '{}'.format(limit)

        offset = '{}'.format(offset)

        return (limit, offset)

    def json_api_call(self, method, path, params):
        """
        Call a Duo API method which is expected to return a JSON body
        with a 200 status. Return the response data structure or raise
        RuntimeError.
        """
        (response, data) = self.api_call(method, path, params)
        return self.parse_json_response(response, data)

    def json_paging_api_call(self, method, path, params):
        """
        Call a Duo API method which is expected to return a JSON body
        with a 200 status. Return a generator that can be used to get
        response data or raise a RuntimeError.
        """
        objects = []
        next_offset = 0

        if 'limit' not in params and self.paging_limit:
            params['limit'] = str(self.paging_limit)

        while next_offset is not None:
            params['offset'] = str(next_offset)
            (response, data) = self.api_call(method, path, params)
            (objects, metadata) = self.parse_json_response_and_metadata(response, data)
            next_offset = metadata.get('next_offset', None)
            for obj in objects:
                yield obj

    def json_cursor_api_call(self, method, path, params, get_records_func):
        """
        Call a Duo API endpoint which utilizes a cursor in some responses to
        page through a set of data. This cursor is supplied through the optional
        "offset" parameter.  The cursor for the next set of data is in the
        response metadata as "next_offset". Callers must also include a
        function parameter to extract the iterable of records to yield. This is
        slightly different than json_paging_api_call because the first request
        does not contain the offset parameter.

        :param method: The method to make the request w/ as a string. Ex:
                       "GET", "POST", "PUT" etc.
        :param path: The path to make the request with as a string.
        :param params: The dict of parameters to send in the request.
        :param get_records_func: Function that can be called to extract an
                                 iterable of records from the parsed response
                                 json.

        :returns: Generator which will yield records from the api response(s).
        """

        next_offset = None

        if 'limit' not in params and self.paging_limit:
            params['limit'] = str(self.paging_limit)

        while True:
            if next_offset is not None:
                params['offset'] = str(next_offset)
            (http_resp, http_resp_data) = self.api_call(method, path, params)
            (response, metadata) = self.parse_json_response_and_metadata(
                http_resp,
                http_resp_data,
            )
            for record in get_records_func(response):
                yield record
            next_offset = metadata.get('next_offset', None)
            if next_offset is None:
                break

    def parse_json_response(self, response, data):
        """
        Return the parsed data structure or raise RuntimeError.
        """
        (response, metadata) = self.parse_json_response_and_metadata(response, data)

        return response

    def parse_json_response_and_metadata(self, response, data):
        """
        Return the parsed data structure and metadata as a tuple or raise RuntimeError.
        """
        def raise_error(msg):
            error = RuntimeError(msg)
            error.status = response.status
            error.reason = response.reason
            error.data = data
            raise error
        if not isinstance(data, str):
            data = data.decode('utf-8')
        if response.status != 200:
            try:
                data = json.loads(data)
                if data['stat'] == 'FAIL':
                    if 'message_detail' in data:
                        raise_error('Received %s %s (%s)' % (
                            response.status,
                            data['message'],
                            data['message_detail'],
                        ))
                    else:
                        raise_error('Received %s %s' % (
                                response.status,
                            data['message'],
                        ))
            except (ValueError, KeyError, TypeError):
                pass
            raise_error('Received %s %s' % (
                    response.status,
                    response.reason,
            ))
        try:
            data = json.loads(data)
            if data['stat'] != 'OK':
                raise_error('Received error response: %s' % data)
            response = data['response']
            metadata = data.get('metadata', {})
            if not metadata and isinstance(response, dict):
                metadata = response.get('metadata', {})

            return (response, metadata)
        except (ValueError, KeyError, TypeError):
            raise_error('Received bad response: %s' % data)

    @classmethod
    def canon_json(cls, params):
        if not isinstance(params, dict):
            raise ValueError('JSON request must be an object.')
        return json.dumps(params, sort_keys=True, separators=(',', ':'))


def output_response(response, data, headers=None):
    """
    Print response, parsed, sorted, and pretty-printed if JSON
    """
    if not headers:
        headers = []
    print(response.status, response.reason)
    for header in headers:
        val = response.getheader(header)
        if val is not None:
            print('%s: %s' % (header, val))
    try:
        if not isinstance(data, str):
            data = data.decode('utf-8')
        data = json.loads(data)
        data = json.dumps(data, sort_keys=True, indent=4)
    except ValueError:
        pass
    print(data)


def main():
    if argparse is None:
        raise argparse_error
    parser = argparse.ArgumentParser()
    # named arguments
    parser.add_argument('--ikey', required=True,
                        help='Duo integration key')
    parser.add_argument('--skey', required=True,
                        help='Duo integration secret key')
    parser.add_argument('--host', required=True,
                        help='Duo API hostname')
    parser.add_argument('--method', required=True,
                        help='HTTP request method')
    parser.add_argument('--path', required=True,
                        help='API endpoint path')
    parser.add_argument('--ca', default=DEFAULT_CA_CERTS)
    parser.add_argument('--sig-version', type=int, default=2)
    parser.add_argument('--sig-timezone', default='UTC')
    parser.add_argument(
        '--show-header',
        action='append',
        default=[],
        metavar='Header-Name',
        help='Show specified response header(s) (default: only output body).',
    )
    parser.add_argument('--file-args', default=[])
    # optional positional arguments are used for GET/POST params, name=val
    parser.add_argument('param', nargs='*')
    args = parser.parse_args()

    client = Client(
        ikey=args.ikey,
        skey=args.skey,
        host=args.host,
        ca_certs=args.ca,
        sig_version=args.sig_version,
        sig_timezone=args.sig_timezone,
    )

    params = collections.defaultdict(list)
    for p in args.param:
        try:
            (k, v) = p.split('=', 1)
        except ValueError:
            sys.exit('Error: Positional argument %s is not '
                     'in key=value format.' % (p,))
        params[k].append(v)

    # parse which arguments are filenames
    file_args = args.file_args
    if args.file_args:
        file_args = file_args.split(',')

    for (k, v) in list(params.items()):
        if k in file_args:      # value is a filename, replace with contents
            if len(v) != 1:
                # file arguments cannot have multiple values
                raise NotImplementedError
            (v,) = v
            with open(v, 'rb') as val:
                params[k] = base64.b64encode(val.read())
        else:
            params[k] = v

    (response, data) = client.api_call(args.method, args.path, params)
    output_response(response, data, args.show_header)

if __name__ == '__main__':
    main()



from .. import util
from .base import TestAdmin
from datetime import datetime, timedelta
from freezegun import freeze_time
import pytz


class TestEndpoints(TestAdmin):

    def test_get_activity_log(self):
        """ Test to get activities log.
        """
        response = self.items_response_client.get_activity_logs(maxtime='1663131599000', mintime='1662958799000')
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v2/logs/activity')
        self.assertEqual(
            util.params_to_dict(args)['account_id'],
            [self.items_response_client.account_id])

    @freeze_time('2022-10-01')
    def test_get_activity_log_with_no_args(self):
        freezed_time = datetime(2022,10,1,0,0,0, tzinfo=pytz.utc)
        expected_mintime = str(int((freezed_time-timedelta(days=180)).timestamp()*1000))
        expected_maxtime = str(int(freezed_time.timestamp() * 1000))
        response = self.items_response_client.get_activity_logs()
        uri, args = response['uri'].split('?')
        param_dict = util.params_to_dict(args)
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v2/logs/activity')
        self.assertEqual(
            param_dict['mintime'],
            [expected_mintime])
        self.assertEqual(
            param_dict['maxtime'],
            [expected_maxtime])



from .. import util
import duo_client.admin
from .base import TestAdmin


MINTIME = 1603399970000
MAXTIME = 1603399973797
LIMIT = 10
NEXT_OFFSET = "99999"
EVENT_TYPE = "auth"


class TestTrustMonitorEvents(TestAdmin):

    def test_get_trust_monitor_events_iterator(self):
        """
        Test to ensure that the correct parameters are supplied when calling
        next on the generator.
        """

        generator = self.client_dtm.get_trust_monitor_events_iterator(
            MINTIME,
            MAXTIME,
            event_type=EVENT_TYPE,
        )
        events = [e for e in generator]
        self.assertEqual(events, [{"foo": "bar"},{"bar": "foo"}])

    def test_get_trust_monitor_events_by_offset(self):
        """
        Test to ensure that the correct parameters are supplied.
        """

        res = self.client_list.get_trust_monitor_events_by_offset(
            MINTIME,
            MAXTIME,
            limit=LIMIT,
            offset=NEXT_OFFSET,
            event_type=EVENT_TYPE,
        )[0]
        uri, qry_str = res["uri"].split("?")
        args = util.params_to_dict(qry_str)

        self.assertEqual(res["method"], "GET")
        self.assertEqual(uri, "/admin/v1/trust_monitor/events")
        self.assertEqual(args["mintime"], ["1603399970000"])
        self.assertEqual(args["maxtime"], ["1603399973797"])
        self.assertEqual(args["offset"], ["99999"])
        self.assertEqual(args["limit"], ["10"])
        self.assertEqual(args["type"], ["auth"])



from .. import util
from .base import TestAdmin


class TestAdminUnits(TestAdmin):
    # Uses underlying paging
    def test_get_administratrive_units(self):
        response = self.client_list.get_administrative_units()
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/administrative_units')

    def test_get_administrative_units_with_limit(self):
        response = self.client_list.get_administrative_units(limit=20)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/administrative_units')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['0'],
            })

    def test_get_adminstrative_units_with_limit_offset(self):
        response = self.client_list.get_administrative_units(limit=20, offset=2)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/administrative_units')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['2'],
            })

    def test_get_administrative_units_with_offset(self):
        response = self.client_list.get_administrative_units(offset=9001)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/administrative_units')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_administrative_units_iterator(self):
        expected_path = '/admin/v1/administrative_units'
        expected_method = 'GET'

        tests = [
            {
                'admin_id': 'aaaa',
                'group_id': '1234',
                'integration_key': 'aaaaaaaaaaaaaaaaaaaa',
            },
            {
                'admin_id': 'aaaa',
                'group_id': '1234',
            },
            {
                'admin_id': 'aaaa',
            },
            {}
        ]

        for test in tests:
            response = (
                self.client_list.get_administrative_units_iterator(**test)
            )
            response = next(response)
            self.assertEqual(response['method'], expected_method)
            (uri, args) = response['uri'].split('?')
            self.assertEqual(uri, expected_path)
            expected_params = {
                key: [value] for (key, value) in test.items()
            }
            expected_params.update(
                {
                    'account_id': [self.client.account_id],
                    'limit': ['100'],
                    'offset': ['0'],
                }
            )
            self.assertEqual(util.params_to_dict(args), expected_params)



#!/usr/bin/env python

""" Script to illustrate how to retrieve a user from the Duo Admin API using the associated email address"""

import sys
import getpass

import duo_client

argv_iter = iter(sys.argv[1:])


def get_next_arg(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    try:
        return next(argv_iter)
    except StopIteration:
        if secure is True:
            return getpass.getpass(prompt)
        else:
            return input(prompt)


def main():
    """ Primary script execution code """
    # Configuration and information about objects to create.
    admin_api = duo_client.Admin(
            ikey=get_next_arg('Admin API integration key ("DI..."): '),
            skey=get_next_arg('integration secret key: ', secure=True),
            host=get_next_arg('API hostname ("api-....duosecurity.com"): '),
    )

    # Retrieve user info from API:
    email_address = get_next_arg('E-mail address of user to retrieve: ')
    user = admin_api.get_user_by_email(email_address)

    if user:
        print(user)
    else:
        print(f"User with email [{email_address}] could not be found.")


if __name__ == '__main__':
    main()



from .. import util
import duo_client.admin
from .base import TestAdmin


class TestPolicies(TestAdmin):
    def setUp(self):
        super(TestPolicies, self).setUp()
        self.client.account_id = None
        self.client_list.account_id = None

    def test_delete_policy_v2(self):
        policy_key = "POSTGY2G0HVWJR4JO1XT"
        response = self.client.delete_policy_v2(policy_key)
        uri, _ = response["uri"].split("?")

        self.assertEqual(response["method"], "DELETE")
        self.assertEqual(uri, "/admin/v2/policies/{}".format(policy_key))

    def test_get_policy_v2(self):
        policy_key = "POSTGY2G0HVWJR4JO1XT"
        response = self.client.get_policy_v2(policy_key)
        uri, _ = response["uri"].split("?")

        self.assertEqual(response["method"], "GET")
        self.assertEqual(uri, "/admin/v2/policies/{}".format(policy_key))

    def test_update_policy_v2(self):
        policy_key = "POSTGY2G0HVWJR4JO1XT"
        policy_settings = {
            "sections": {
                "browsers": {
                    "blocked_browsers_list": ["ie"],
                }
            }
        }
        response = self.client.update_policy_v2(policy_key, policy_settings)

        self.assertEqual(response["method"], "PUT")
        self.assertEqual(response["uri"], "/admin/v2/policies/{}".format(policy_key))

    def test_update_policies_v2(self):
        edit_list = ["POSTGY2G0HVWJR4JO1XT", "POSTGY2G0HVWJR4JO1XU"]
        sections = {
                "browsers": {
                    "blocked_browsers_list": ["ie"],
                }
        }
        response = self.client.update_policies_v2(sections, [], edit_list)
        self.assertEqual(response["method"], "PUT")
        self.assertEqual(response["uri"], "/admin/v2/policies/update")

    def test_create_policy_v2(self):
        policy_settings = {
            "name": "my new policy",
            "sections": {
                "browsers": {
                    "blocked_browsers_list": ["ie"],
                }
            },
        }
        response = self.client.create_policy_v2(policy_settings)

        self.assertEqual(response["method"], "POST")
        self.assertEqual(response["uri"], "/admin/v2/policies")

    def test_copy_policy_v2(self):
        policy_key = "POSTGY2G0HVWJR4JO1XT"
        new_policy_names_list = ["Copied Pol 1", "Copied Pol 2"]
        response = self.client.copy_policy_v2(policy_key, new_policy_names_list)

        self.assertEqual(response["method"], "POST")
        self.assertEqual(response["uri"], "/admin/v2/policies/copy")

    def test_get_policies_v2(self):
        response = self.client.get_policies_v2(limit=3, offset=0)
        uri, args = response["uri"].split("?")

        self.assertEqual(response["method"], "GET")
        self.assertEqual(uri, "/admin/v2/policies")
        self.assertDictEqual(
            util.params_to_dict(args), {"limit": ["3"], "offset": ["0"]}
        )

    def test_get_policies_v2_iterator(self):
        iterator = self.client_list.get_policies_v2_iterator()
        response = next(iterator)
        uri, args = response["uri"].split("?")

        self.assertEqual(response["method"], "GET")
        self.assertEqual(uri, "/admin/v2/policies")
        self.assertDictEqual(
            util.params_to_dict(args), {"limit": ["100"], "offset": ["0"]}
        )

    def test_get_policy_summary(self):
        response = self.client.get_policy_summary_v2()
        uri, _ = response["uri"].split("?")

        self.assertEqual(response["method"], "GET")
        self.assertEqual(uri, "/admin/v2/policies/summary")

    def test_calculate_policy(self):
        ikey = "DI82WWNVI5Z4V10LZJR6"
        ukey = "DUQU89MDEWOUR277H44G"

        response = self.client.calculate_policy(integration_key=ikey, user_id=ukey)
        uri, args = response["uri"].split("?")

        self.assertEqual(response["method"], "GET")
        self.assertEqual(uri, "/admin/v2/policies/calculate")
        self.assertDictEqual(
            util.params_to_dict(args), {"integration_key": [ikey], "user_id": [ukey]}
        )


from .. import util
import duo_client.admin
from .base import TestAdmin


class TestEndpoints(TestAdmin):
    def test_get_authentication_log_v1(self):
        """ Test to get authentication log on version 1 api.
        """
        response = self.client_list.get_authentication_log(api_version=1)[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/logs/authentication')
        self.assertEqual(
            util.params_to_dict(args)['account_id'],
            [self.client_list.account_id])

    def test_get_authentication_log_v2(self):
        """ Test to get authentication log on version 1 api.
        """
        response = self.client_authlog.get_authentication_log(api_version=2)
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v2/logs/authentication')
        self.assertEqual(
            util.params_to_dict(args)['account_id'],
            [self.client_authlog.account_id])



#!/usr/bin/env python
"""Print Duo Trust Monitor Events which surfaced within the past two weeks."""

import json
import sys
from datetime import datetime, timedelta, timezone

from duo_client import Admin

argv_iter = iter(sys.argv[1:])
def get_next_arg(prompt):
    try:
        return next(argv_iter)
    except StopIteration:
        return input(prompt)


def main(args):
    # Instantiate the Admin client object.
    admin_client = Admin(args[0], args[1], args[2])

    # Query for Duo Trust Monitor events that were surfaced within the last two weeks (from today).
    now = datetime.now(tz=timezone.utc)
    mintime_ms = int((now - timedelta(weeks=2)).timestamp() * 1000)
    maxtime_ms = int(now.timestamp() * 1000)

    # Loop over the returned iterator to navigate through each event, printing it to stdout.
    for event in admin_client.get_trust_monitor_events_iterator(mintime_ms, maxtime_ms):
        print(json.dumps(event, sort_keys=True))


def parse_args():
    ikey=get_next_arg('Duo Admin API integration key ("DI..."): ')
    skey=get_next_arg('Duo Admin API integration secret key: ')
    host=get_next_arg('Duo Admin API hostname ("api-....duosecurity.com"): ')
    return (ikey, skey, host,)


if __name__ == "__main__":
    args = parse_args()
    main(args)



#!/usr/bin/env python
import csv
import sys
import duo_client

argv_iter = iter(sys.argv[1:])
def get_next_arg(prompt):
    try:
        return next(argv_iter)
    except StopIteration:
        return input(prompt)

# Configuration and information about objects to create.
admin_api = duo_client.Admin(
    ikey=get_next_arg('Admin API integration key ("DI..."): '),
    skey=get_next_arg('integration secret key: '),
    host=get_next_arg('API hostname ("api-....duosecurity.com"): '),
)

# Retrieve log info from API:
logs = admin_api.get_authentication_log()

# Count authentications by country:
counts = dict()
for log in logs:
    country = log['location']['country']
    if country != '':
        counts[country] = counts.get(country, 0) + 1

# Print CSV of country, auth count:
auths_descending = sorted(counts.items(), reverse=True)
reporter = csv.writer(sys.stdout)
print("[+] Report of auth counts by country:")
reporter.writerow(('Country', 'Auth Count'))
for row in auths_descending:
    reporter.writerow([
        row[0],
        row[1],
    ])



#!/usr/bin/env python
import csv
import sys

import duo_client

argv_iter = iter(sys.argv[1:])
def get_next_arg(prompt):
    try:
        return next(argv_iter)
    except StopIteration:
        return input(prompt)

# Configuration and information about objects to create.
admin_api = duo_client.Admin(
    ikey=get_next_arg('Admin API integration key ("DI..."): '),
    skey=get_next_arg('integration secret key: '),
    host=get_next_arg('API hostname ("api-....duosecurity.com"): '),
)

# Retrieve user info from API:
users = admin_api.get_users()

# Print CSV of username, phone number, phone type, and phone platform:
#
# (If a user has multiple phones, there will be one line printed per
# associated phone.)
reporter = csv.writer(sys.stdout)
print("[+] Report of all users and associated phones:")
reporter.writerow(('Username', 'Phone Number', 'Type', 'Platform'))
for user in users:
    for phone in user["phones"]:
        reporter.writerow([
                user["username"],
                phone["number"],
                phone["type"],
                phone["platform"],
        ])



from .. import util
import duo_client.admin
from .base import TestAdmin


class TestWebauthn(TestAdmin):
    def test_get_webauthncredentials_with_params(self):
        """ Test to get webauthn credentials with params.
        """
        response = list(self.client_list.get_webauthncredentials(limit=8))[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/webauthncredentials')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id':[self.client_list.account_id],
                'limit': ['8'],
                'offset': ['0'],
            }
        )

    def test_get_webauthncredentials_iterator(self):
        response = self.client_list.get_webauthncredentials_iterator()
        response = next(response)
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/webauthncredentials')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0']
            }
        )

    def test_get_webauthncredentials_without_params(self):
        """ Test to get webauthn credentials without params.
        """
        response = list(self.client_list.get_webauthncredentials())[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/webauthncredentials')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
             }
        )

    def test_get_webauthncredentials_with_offset(self):
        response = list(self.client_list.get_webauthncredentials(limit=2, offset=3))[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/webauthncredentials')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id':[self.client_list.account_id],
                'limit': ['2'],
                'offset': ['3']
            }
        )

    def test_get_webauthncredential_by_id(self):
        """ Test to get webauthn credential by registration id.
        """
        response = self.client.get_webauthncredential_by_id('DU012345678901234567')
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/webauthncredentials/DU012345678901234567')
        self.assertEqual(util.params_to_dict(args),
                         {'account_id':[self.client.account_id]})

    def test_delete_webauthncredential(self):
        """ Test to delete webauthn credential by registration id.
        """
        response = self.client.delete_webauthncredential('DU012345678901234567')
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'DELETE')
        self.assertEqual(uri, '/admin/v1/webauthncredentials/DU012345678901234567')
        self.assertEqual(util.params_to_dict(args),
                         {'account_id':[self.client.account_id]})


import json

from .base import TestAdmin
from .. import util


class TestPassport(TestAdmin):
    def test_get_passport(self):
        """ Test get passport configuration
        """
        response = self.client.get_passport_config()
        (uri, args) = response['uri'].split('?')
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v2/passport/config')
        self.assertEqual(util.params_to_dict(args), {'account_id': [self.client.account_id]})

    def test_update_passport(self):
        """ Test update passport configuration
        """
        response = self.client.update_passport_config(enabled_status="enabled-for-groups", enabled_groups=["passport-test-group"])
        self.assertEqual(response["uri"], "/admin/v2/passport/config")
        body = json.loads(response["body"])
        self.assertEqual(body["enabled_status"], "enabled-for-groups")
        self.assertEqual(body["enabled_groups"], ["passport-test-group"])
        self.assertEqual(body["disabled_groups"], [])



"""
Example of Duo Auth API uaer authentication with synchronous request/response
"""

import duo_client
import sys
import getpass

from pprint import pprint


argv_iter = iter(sys.argv[1:])


def _get_next_arg(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    try:
        return next(argv_iter)
    except StopIteration:
        if secure is True:
            return getpass.getpass(prompt)
        else:
            return input(prompt)


def prompt_for_credentials() -> dict:
    """Collect required API credentials from command line prompts

    :return: dictionary containing Duo Auth API ikey, skey and hostname strings
    """

    ikey = _get_next_arg('Duo Auth API integration key ("DI..."): ')
    skey = _get_next_arg('Duo Auth API integration secret key: ', secure=True)
    host = _get_next_arg('Duo Auth API hostname ("api-....duosecurity.com"): ')
    username = _get_next_arg('Duo Username: ')

    return {"USERNAME": username, "IKEY": ikey, "SKEY": skey, "APIHOST": host}


def main():
    """Main program entry point"""

    inputs = prompt_for_credentials()

    auth_client = duo_client.Auth(
            ikey=inputs['IKEY'],
            skey=inputs['SKEY'],
            host=inputs['APIHOST']
    )

    # Verify that the Duo service is available
    duo_ping = auth_client.ping()
    if 'time' in duo_ping:
        print("\nDuo service check completed successfully.")
    else:
        print(f"Error: {duo_ping}")

    # Verify that IKEY and SKEY information provided are valid
    duo_check= auth_client.check()
    if 'time' in duo_check:
        print("IKEY and SKEY provided have been verified.")
    else:
        print(f"Error: {duo_check}")

    # Execute pre-authentication for given user
    print(f"\nExecuting pre-authentication for {inputs['USERNAME']}...")
    pre_auth = auth_client.preauth(username=inputs['USERNAME'])

    if pre_auth['result'] == "auth":
        try:
            # User exists and has an MFA device enrolled
            print(f"Executing authentication action for {inputs['USERNAME']}...")
            # "auto" is selected for the factor in this example, however the pre_auth['devices'] dictionary
            # element contains a list of factors available for the provided user, if an alternate method is desired
            auth = auth_client.auth(factor="auto", username=inputs['USERNAME'], device="auto")
            print(f"\n{auth['status_msg']}")
        except Exception as e_str:
            print(e_str)
    elif pre_auth['result'] == "allow":
        # User is in bypass mode
        print(pre_auth['status_msg'])
    elif pre_auth['result'] == "enroll":
        # User is unknown and not enrolled in Duo with a 'New User' policy setting of 'Require Enrollment'
        # Setting a 'New User' policy to 'Require Enrollment' should only be done for Group level policies where
        # the intent is to capture "partially enrolled" users. "Parially enrolled" users are those that Duo has a
        # defined username but does not have an MFA device enrolled.
        print("Please enroll in Duo using the following URL.")
        print(pre_auth['enroll_portal_url'])
    elif pre_auth['result'] == "deny":
        # User is denied by policy setting
        print(pre_auth['status_msg'])
    else:
        print("Error: an unexpected error occurred")
        print(pre_auth)


if __name__ == '__main__':
    main()



import unittest

from .. import util
import json
import duo_client.admin
from .base import TestAdmin


class TestIntegration(TestAdmin):
    def setUp(self):
        super(TestIntegration, self).setUp()
        self.integration_key = "DISRYL7L8LZ5YXNWKGNK"

    def test_get_integration(self):
        response = self.client.get_integration(self.integration_key)
        (uri, args) = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v3/integrations/{}'.format(self.integration_key))
        self.assertEqual(util.params_to_dict(args), {'account_id': [self.client.account_id]})

    def test_delete_integration(self):
        response = self.client.delete_integration(self.integration_key)
        (uri, args) = response['uri'].split('?')

        self.assertEqual(response['method'], 'DELETE')
        self.assertEqual(uri, '/admin/v3/integrations/{}'.format(self.integration_key))
        self.assertEqual(util.params_to_dict(args), {'account_id': [self.client.account_id]})

    def test_create_integration(self):
        response = self.client.create_integration(
            name="New integration name",
            integration_type="sso-generic",
            sso={
                "idp_metadata": None,
                "saml_config": {}
            },
        )

        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/admin/v3/integrations')
        self.assertEqual(json.loads(response['body']),
            {
                "account_id": self.client.account_id,
                "name": "New integration name",
                "type": "sso-generic",
                "sso": {
                    "idp_metadata": None,
                    "saml_config": {}
                },
            }
        )

    def test_update_integration_success(self):
        response = self.client.update_integration(
            self.integration_key,
            name="Integration name",
            sso={
                "saml_config": {
                    "nameid_attribute": "mail",
                }
            },
        )

        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/admin/v3/integrations/{}'.format(self.integration_key))
        self.assertEqual(json.loads(response['body']),
            {
                "account_id": self.client.account_id,
                "name": "Integration name",
                "sso": {
                    "saml_config": {
                        "nameid_attribute": "mail",
                    }
                },
            }
        )

if __name__ == '__main__':
    unittest.main()



import json
from .. import util
import duo_client.admin
from .base import TestAdmin


class TestAdmins(TestAdmin):
    # Uses underlying paging
    def test_get_admins(self):
        response = self.client_list.get_admins()
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_admins_with_limit(self):
        response = self.client_list.get_admins(limit='20')
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['0'],
            })

    def test_get_admins_with_limit_offset(self):
        response = self.client_list.get_admins(limit='20', offset='2')
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['2'],
            })

    def test_get_admins_with_offset(self):
        response = self.client_list.get_admins(offset=9001)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_admins_iterator(self):
        response = self.client_list.get_admins_iterator()
        response = next(response)
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_external_password_mgmt_statuses(self):

        response = self.client_list.get_external_password_mgmt_statuses()
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins/password_mgmt')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_external_password_mgmt_statuses_with_limit(self):
        response = self.client_list.get_external_password_mgmt_statuses(
            limit='20')
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins/password_mgmt')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['0'],
            })

    def test_get_external_password_mgmt_statusesg_with_limit_offset(self):
        response = self.client_list.get_external_password_mgmt_statuses(
            limit='20', offset='2')
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins/password_mgmt')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['2'],
            })

    def test_get_external_password_mgmt_statuses_with_offset(self):
        response = self.client_list.get_external_password_mgmt_statuses(
            offset=9001)
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins/password_mgmt')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_external_password_mgmt_status_for_admin(self):
        response = self.client_list.get_external_password_mgmt_status_for_admin(
            'DFAKEADMINID')
        response = response[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/admins/DFAKEADMINID/password_mgmt')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
            })

    def test_update_admin_password_mgmt_status(self):
        response = self.client_list.update_admin_password_mgmt_status(
            'DFAKEADMINID', has_external_password_mgmt='False')
        response = response[0]
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/admin/v1/admins/DFAKEADMINID/password_mgmt')
        self.assertEqual(
            json.loads(response['body']),
            {
                'account_id': self.client.account_id,
                'has_external_password_mgmt': 'False'
            })

    def test_update_admin_password_mgmt_status_set_password(self):
        response = self.client_list.update_admin_password_mgmt_status(
            'DFAKEADMINID', has_external_password_mgmt='True', password='dolphins')
        response = response[0]
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/admin/v1/admins/DFAKEADMINID/password_mgmt')
        self.assertEqual(
            json.loads(response['body']),
            {
                'account_id': self.client.account_id,
                'has_external_password_mgmt': 'True',
                'password': 'dolphins'
            })



"""
Example of Duo Auth API user authentication using asynchronous resquest/response methods
"""

import duo_client
import sys
import getpass


argv_iter = iter(sys.argv[1:])


def _get_next_arg(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    try:
        return next(argv_iter)
    except StopIteration:
        if secure is True:
            return getpass.getpass(prompt)
        else:
            return input(prompt)


def prompt_for_credentials() -> dict:
    """Collect required API credentials from command line prompts

    :return: dictionary containing Duo Auth API ikey, skey and hostname strings
    """

    ikey = _get_next_arg('Duo Auth API integration key ("DI..."): ')
    skey = _get_next_arg('Duo Auth API integration secret key: ', secure=True)
    host = _get_next_arg('Duo Auth API hostname ("api-....duosecurity.com"): ')
    username = _get_next_arg('Duo Username: ')

    return {"USERNAME": username, "IKEY": ikey, "SKEY": skey, "APIHOST": host}


def main():
    """Main program entry point"""

    inputs = prompt_for_credentials()

    auth_client = duo_client.Auth(
            ikey=inputs['IKEY'],
            skey=inputs['SKEY'],
            host=inputs['APIHOST']
    )

    # Verify that the Duo service is available
    duo_ping = auth_client.ping()
    if 'time' in duo_ping:
        print("\nDuo service check completed successfully.")
    else:
        print(f"Error: {duo_ping}")

    # Verify that IKEY and SKEY information provided are valid
    duo_check= auth_client.check()
    if 'time' in duo_check:
        print("IKEY and SKEY provided have been verified.")
    else:
        print(f"Error: {duo_check}")

    # Execute pre-authentication for given user
    print(f"\nExecuting pre-authentication for {inputs['USERNAME']}...")
    pre_auth = auth_client.preauth(username=inputs['USERNAME'])

    if pre_auth['result'] == "auth":
        try:
            print(f"Executing authentication action for {inputs['USERNAME']}...")
            auth = auth_client.auth(factor="push", username=inputs['USERNAME'], device="auto", async_txn=True)
            if 'txid' in auth:
                waiting = True
                # Collect the authentication result
                print("Getting authentication result...")
                # Repeat long polling for async authentication status until no longer in a 'waiting' state
                while waiting is True:
                    # Poll Duo Auth API for the status of the async authentication based upon transaction ID
                    auth_status = auth_client.auth_status(auth['txid'])
                    print(f"Auth status: {auth_status}")
                    if auth_status['waiting'] is not True:
                        # Waiting for response too async authentication is no longer 'True', so break the loop
                        waiting = False
                # Parse response for the 'status' dictionary key to determine whether to allow or deny
                print(auth_status)
            else:
                # Some kind of unexpected error occurred
                print(f"Error: an unknown error occurred attempting authentication for [{inputs['USERNAME']}]")
        except Exception as e_str:
            print(e_str)
    else:
        print(pre_auth['status_msg'])


if __name__ == '__main__':
    main()



import json

from .. import util
import duo_client.admin
from .base import TestAccountAdmin


class TestBilling(TestAccountAdmin):
    def test_get_billing_edition(self):
        """Test to get billing edition
        """
        response = self.client.get_edition()
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/billing/edition')
        self.assertEqual(util.params_to_dict(args),
                         {
                             'account_id': [self.client.account_id],
                         })

    def test_set_business_billing_edition(self):
        """Test to set PLATFORM billing edition
        """
        response = self.client.set_edition('PLATFORM')
        uri = response['uri']

        self.assertEqual(response['method'], 'POST')
        self.assertEqual(uri, '/admin/v1/billing/edition')
        self.assertEqual(json.loads(response['body']),
                         {
                             'edition': 'PLATFORM',
                             'account_id': self.client.account_id,
                         })

    def test_set_enterprise_billing_edition(self):
        """Test to set ENTERPRISE billing edition
        """
        response = self.client.set_edition('ENTERPRISE')
        uri = response['uri']

        self.assertEqual(response['method'], 'POST')
        self.assertEqual(uri, '/admin/v1/billing/edition')
        self.assertEqual(json.loads(response['body']),
                         {
                             'edition': 'ENTERPRISE',
                             'account_id': self.client.account_id,
                         })

    def test_get_telephony_credits(self):
        """Test to get telephony credits
        """
        response = self.client.get_telephony_credits()
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/billing/telephony_credits')
        self.assertEqual(util.params_to_dict(args),
                         {
                             'account_id': [self.client.account_id],
                         })

    def test_set_telephony_credits(self):
        """Test to set telephony credits
        """
        response = self.client.set_telephony_credits(10)
        uri = response['uri']

        self.assertEqual(response['method'], 'POST')
        self.assertEqual(uri, '/admin/v1/billing/telephony_credits')
        self.assertEqual(json.loads(response['body']),
                         {
                             'credits': '10',
                             'account_id': self.client.account_id,
                         })



#!/usr/bin/python
import pprint
import sys

import duo_client

argv_iter = iter(sys.argv[1:])


def get_next_arg(prompt):
    try:
        return next(argv_iter)
    except StopIteration:
        return input(prompt)


ikey = get_next_arg('Admin API integration key ("DI..."): ')
skey = get_next_arg('integration secret key: ')
host = get_next_arg('API hostname ("api-....duosecurity.com"): ')

# Configuration and information about objects to create.
admin_api = duo_client.Admin(
        ikey,
        skey,
        host,
)

integration = admin_api.create_integration(
        name='api-created integration',
        integration_type='sso-generic',
        sso={
                "saml_config": {
                        "entity_id":                 "entity_id",
                        "acs_urls":                  [
                                {
                                        "url":       "https://example.com/acs",
                                        "binding":   None,
                                        "isDefault": None,
                                        "index":     None,
                                }
                        ],
                        "nameid_format":             "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                        "nameid_attribute":          "mail",
                        "sign_assertion":            False,
                        "sign_response":             True,
                        "signing_algorithm":         "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
                        "mapped_attrs":              {},
                        "relaystate":                "https://example.com/relaystate",
                        "slo_url":                   "https://example.com/slo",
                        "spinitiated_url":           "https://example.com/spurl",
                        "static_attrs":              {},
                        "role_attrs":                {
                                "bob": {
                                        "ted":   ["DGS08MMO53GNRLSFW0D0", "DGETXINZ6CSJO4LRSVKV"],
                                        "frank": ["DGETXINZ6CSJO4LRSVKV"],
                                }
                        },
                        "attribute_transformations": {
                                "attribute_1": 'use "<Username>"\nprepend text="dev-"',
                                "attribute_2": 'use "<Email Address>"\nappend additional_attr="<First Name>"',
                        }
                }
        },
)

print('Created integration:')
pprint.pprint(integration)



"""
Example of Duo Accounts API set child account edition
"""

import duo_client
import getpass

ALLOWED_DUO_EDITIONS = ("PERSONAL", "ENTERPRISE", "PLATFORM", "BEYOND")

def _get_user_input(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    if secure is True:
        return getpass.getpass(prompt)
    else:
        return input(prompt)


def prompt_for_credentials() -> dict:
    """Collect required API credentials from command line prompts"""

    ikey = _get_user_input('Duo Accounts API integration key ("DI..."): ')
    skey = _get_user_input('Duo Accounts API integration secret key: ', secure=True)
    host = _get_user_input('Duo Accounts API hostname ("api-....duosecurity.com"): ')
    account_id = _get_user_input('Child account ID: ')
    account_apihost = _get_user_input('Child account api_hostname: ')
    account_edition = _get_user_input('Child account edition: ')
    while account_edition.upper() not in ALLOWED_DUO_EDITIONS:
        print(f"Invalid account edition. Please select one of {ALLOWED_DUO_EDITIONS}")
        account_edition = _get_user_input('Child account edition: ')

    return {
            "ikey": ikey,
            "skey": skey,
            "host": host,
            "account_id": account_id,
            "child_api_host": account_apihost,
            "account_edition": account_edition,
    }


def main():
    """Main program entry point"""

    inputs = prompt_for_credentials()
    edition = inputs.pop('account_edition')
    edition = edition.upper()

    account_admin_api = duo_client.admin.AccountAdmin(**inputs)

    print(f"Setting edition for account ID {inputs['account_id']} to {edition}")
    result = account_admin_api.set_edition(edition)
    if result != "":
        print(f"An error occurred while setting edition for account {inputs['account_id']}")
        print(f"Error message: {result}")
    else:
        print(f"Edition [{edition}] successfully set for account ID {inputs['account_id']}")


if __name__ == '__main__':
    main()



import unittest

from .. import util
import duo_client.admin
from .base import TestAdmin


class TestUserPhones(TestAdmin):
    def test_get_user_phones_iterator(self):
        """Test to get phones iterator by user id
        """
        iterator = self.client_list.get_user_phones_iterator(
            'DU012345678901234567')
        response = next(iterator)
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users/DU012345678901234567/phones')
        self.assertEqual(util.params_to_dict(args),
                         {
                            'account_id':[self.client.account_id],
                            'limit': ['100'],
                            'offset': ['0'],
                        })

    def test_get_user_phones(self):
        """Test to get phones by user id
        """
        response = self.client_list.get_user_phones('DU012345678901234567')[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users/DU012345678901234567/phones')
        self.assertEqual(util.params_to_dict(args),
                         {
                            'account_id':[self.client.account_id],
                            'limit': ['100'],
                            'offset': ['0'],
                        })

    def test_get_user_phones_with_offset(self):
        """Test to get phones by user id with pagination params
        """
        response = self.client_list.get_user_phones(
            'DU012345678901234567', offset=30)[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users/DU012345678901234567/phones')
        self.assertEqual(util.params_to_dict(args),
                         {
                            'account_id':[self.client.account_id],
                            'limit': ['100'],
                            'offset': ['0'],
                         })

    def test_get_user_phones_with_limit(self):
        """Test to get phones by user id with pagination params
        """
        response = self.client_list.get_user_phones(
            'DU012345678901234567', limit=10)[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users/DU012345678901234567/phones')
        self.assertEqual(util.params_to_dict(args),
                         {
                            'account_id':[self.client.account_id],
                            'limit': ['10'],
                            'offset': ['0'],
                         })

    def test_get_user_phones_with_limit_and_offset(self):
        """Test to get phones by user id with pagination params
        """
        response = self.client_list.get_user_phones(
            'DU012345678901234567', limit=10, offset=30)[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users/DU012345678901234567/phones')
        self.assertEqual(util.params_to_dict(args),
                         {
                            'account_id':[self.client.account_id],
                            'limit': ['10'],
                            'offset': ['30'],
                         })


if __name__ == '__main':
    unittest.main()



from .. import util
from .base import TestAdmin
from datetime import datetime, timedelta
from freezegun import freeze_time
import pytz


class TestTelephonyLogEndpoints(TestAdmin):
    def test_get_telephony_logs_v2(self):
        """Test to get activities log."""
        response = self.items_response_client.get_telephony_log(
            maxtime=1663131599000, mintime=1662958799000, api_version=2
        )
        uri, args = response["uri"].split("?")

        self.assertEqual(response["method"], "GET")
        self.assertEqual(uri, "/admin/v2/logs/telephony")
        self.assertEqual(
            util.params_to_dict(args)["account_id"], [self.items_response_client.account_id]
        )

    @freeze_time("2022-10-01")
    def test_get_telephony_logs_v2_no_args(self):
        freezed_time = datetime(2022, 10, 1, 0, 0, 0, tzinfo=pytz.utc)
        expected_mintime = str(
            int((freezed_time - timedelta(days=180)).timestamp() * 1000)
        )
        expected_maxtime = str(int(freezed_time.timestamp() * 1000) - 120)
        response = self.items_response_client.get_telephony_log(api_version=2)
        uri, args = response["uri"].split("?")
        param_dict = util.params_to_dict(args)
        self.assertEqual(response["method"], "GET")
        self.assertEqual(uri, "/admin/v2/logs/telephony")
        self.assertEqual(param_dict["mintime"], [expected_mintime])
        self.assertEqual(param_dict["maxtime"], [expected_maxtime])

    @freeze_time("2022-10-01")
    def test_get_telephony_logs_v1_no_args(self):
        response = self.client_list.get_telephony_log()
        uri, args = response[0]["uri"].split("?")
        self.assertEqual(response[0]["method"], "GET")
        self.assertEqual(uri, "/admin/v1/logs/telephony")



"""
Example of Duo Accounts API child account creation
"""

import duo_client

import sys
import getpass

from pprint import pprint


argv_iter = iter(sys.argv[1:])


def _get_next_arg(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    try:
        return next(argv_iter)
    except StopIteration:
        if secure is True:
            return getpass.getpass(prompt)
        else:
            return input(prompt)


def prompt_for_credentials() -> dict:
    """Collect required API credentials from command line prompts

    :return: dictionary containing Duo Accounts API ikey, skey and hostname strings
    """

    ikey = _get_next_arg('Duo Accounts API integration key ("DI..."): ')
    skey = _get_next_arg('Duo Accounts API integration secret key: ', secure=True)
    host = _get_next_arg('Duo Accounts API hostname ("api-....duosecurity.com"): ')
    account_name = _get_next_arg('Name for new child account: ')

    return {"IKEY": ikey, "SKEY": skey, "APIHOST": host, "ACCOUNT_NAME": account_name}


def main():
    """Main program entry point"""

    inputs = prompt_for_credentials()

    account_client = duo_client.Accounts(
            ikey=inputs['IKEY'],
            skey=inputs['SKEY'],
            host=inputs['APIHOST']
    )

    print(f"Creating child account with name [{inputs['ACCOUNT_NAME']}]")
    child_account = account_client.create_account(inputs['ACCOUNT_NAME'])

    if 'account_id' in child_account:
        print(f"Child account for [{inputs['ACCOUNT_NAME']}] created successfully.")
    else:
        print(f"An unexpected error occurred while creating child account for {inputs['ACCOUNT_NAME']}")
    print(child_account)


if __name__ == '__main__':
    main()



"""
Example of Duo account API uaer accountentication with synchronous request/response
"""

import duo_client
import sys
import getpass

from pprint import pprint


argv_iter = iter(sys.argv[1:])


def _get_next_arg(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    try:
        return next(argv_iter)
    except StopIteration:
        if secure is True:
            return getpass.getpass(prompt)
        else:
            return input(prompt)


def prompt_for_credentials() -> dict:
    """Collect required API credentials from command line prompts

    :return: dictionary containing Duo Accounts API ikey, skey and hostname strings
    """

    ikey = _get_next_arg('Duo Accounts API integration key ("DI..."): ')
    skey = _get_next_arg('Duo Accounts API integration secret key: ', secure=True)
    host = _get_next_arg('Duo Accounts API hostname ("api-....duosecurity.com"): ')

    return {"IKEY": ikey, "SKEY": skey, "APIHOST": host}


def main():
    """Main program entry point"""

    inputs = prompt_for_credentials()

    account_client = duo_client.Accounts(
            ikey=inputs['IKEY'],
            skey=inputs['SKEY'],
            host=inputs['APIHOST']
    )

    child_accounts = account_client.get_child_accounts()

    if isinstance(child_accounts, list):
        # Expected list of child accounts returned
        for child_account in child_accounts:
            print(child_account)

    if isinstance(child_accounts, dict):
        # Non-successful response returned
        print(child_accounts)


if __name__ == '__main__':
    main()



import duo_client.admin
from .. import util
from .base import TestAdmin


class TestBypassCodes(TestAdmin):
    def test_delete_bypass_code_by_id(self):
        """ Test to delete a bypass code by id.
        """
        response = self.client.delete_bypass_code_by_id('DU012345678901234567')
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'DELETE')
        self.assertEqual(uri, '/admin/v1/bypass_codes/DU012345678901234567')
        self.assertEqual(util.params_to_dict(args),
                         {'account_id': [self.client.account_id]})

    def test_get_bypass_codes_generator(self):
        """ Test to get bypass codes generator.
        """
        generator = self.client_list.get_bypass_codes_generator()
        response = next(generator)
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/bypass_codes')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
             })

    def test_get_bypass_codes(self):
        """ Test to get bypass codes without params.
        """
        response = self.client_list.get_bypass_codes()[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/bypass_codes')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
             })

    def test_get_bypass_codes_limit(self):
        """ Test to get bypass codes with limit.
        """
        response = self.client_list.get_bypass_codes(limit='20')[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/bypass_codes')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['20'],
                'offset': ['0'],
            })

    def test_get_bypass_codes_offset(self):
        """ Test to get bypass codes with offset.
        """
        response = self.client_list.get_bypass_codes(offset='20')[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/bypass_codes')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_bypass_codes_limit_offset(self):
        """ Test to get bypass codes with limit and offset.
        """
        response = self.client_list.get_bypass_codes(limit='20', offset='2')[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/bypass_codes')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['20'],
                'offset': ['2'],
            })



import json
from .base import TestAdmin
import os
import base64
import urllib.parse


class TestLogo(TestAdmin):

    def test_update_logo(self):
        # Get an image to upload:
        logo_path = os.path.join(self.TEST_RESOURCES_DIR, "barn-owl-small.png")
        with open(logo_path, "rb") as image_file:
            logo_file = image_file.read()

            # Call function in client:
            response = self.client.update_logo(logo_file)

            # Prep validation text:
            base64_logo = base64.b64encode(logo_file)
            base64_logo = urllib.parse.quote_plus(base64_logo)

            # Validate response:
            self.assertTrue(
                json.loads(response['body']).get('logo'),
                base64_logo
            )



#!/usr/bin/env python
import sys

import duo_client

EDITIONS = {
        "ENTERPRISE": "Duo Essentials",
        "PLATFORM": "Duo Advantage",
        "BEYOND": "Duo Premier",
        "PERSONAL": "Duo Free"
}

def get_next_input(prompt):
    try:
        return next(iter(sys.argv[1:]))
    except StopIteration:
        return input(prompt)


def main():
    """Program entry point"""
    ikey=get_next_input('Accounts API integration key ("DI..."): ')
    skey=get_next_input('Accounts API integration secret key: ')
    host=get_next_input('Accounts API hostname ("api-....duosecurity.com"): ')

    # Configuration and information about objects to create.
    accounts_api = duo_client.Accounts(
        ikey=ikey,
        skey=skey,
        host=host,
    )

    kwargs = {
            'ikey': ikey,
            'skey': skey,
            'host': host,
    }

    # Get all child accounts
    child_accounts = accounts_api.get_child_accounts()

    for child_account in child_accounts:
        # Create AccountAdmin with child account_id, child api_hostname and kwargs consisting of ikey, skey, and host
        account_admin_api = duo_client.admin.AccountAdmin(
                child_account['account_id'],
                child_api_host = child_account['api_hostname'],
                **kwargs,
        )
        try:
            # Get edition of child account
            child_account_edition = account_admin_api.get_edition()
            print(f"Edition for child account {child_account['name']}: {child_account_edition['edition']}")
        except RuntimeError as err:
            # The account might not have access to get billing information
            if "Received 403 Access forbidden" == str(err):
                print("{error}: No access for billing feature".format(error=err))
            else:
                print(err)

        try:
            # Get telephony credits of child account
            child_telephony_credits = account_admin_api.get_telephony_credits()
            print("Telephony credits for child account {name}: {edition}".format(
                name=child_account['name'],
                edition=child_telephony_credits['credits'])
            )
        except RuntimeError as err:
            # The account might not have access to get telephony credits
            if "Received 403 Access forbidden" == str(err):
                print("{error}: No access for telephony feature".format(error=err))
            else:
                print(err)


if __name__ == "__main__":
    main()



#!/usr/bin/python
import pprint
import sys

import duo_client

argv_iter = iter(sys.argv[1:])
def get_next_arg(prompt):
    try:
        return next(argv_iter)
    except StopIteration:
        return input(prompt)

# Configuration and information about objects to create.
admin_api = duo_client.Admin(
    ikey=get_next_arg('Admin API integration key ("DI..."): '),
    skey=get_next_arg('integration secret key: '),
    host=get_next_arg('API hostname ("api-....duosecurity.com"): '),
)

USERNAME = get_next_arg('user login name: ')
REALNAME = get_next_arg('user full name: ')

# Refer to http://www.duosecurity.com/docs/adminapi for more
# information about phone types and platforms.
PHONE_NUMBER = get_next_arg('phone number (e.g. +1-555-123-4567): ')
PHONE_TYPE = get_next_arg('phone type (e.g. mobile): ')
PHONE_PLATFORM = get_next_arg('phone platform (e.g. google android): ')

# Create and return a new user object.
user = admin_api.add_user(
    username=USERNAME,
    realname=REALNAME,
)
print('Created user:')
pprint.pprint(user)

# Create and return a new phone object.
phone = admin_api.add_phone(
    number=PHONE_NUMBER,
    type=PHONE_TYPE,
    platform=PHONE_PLATFORM,
)
print('Created phone:')
pprint.pprint(phone)

# Associate the user with the phone.
admin_api.add_user_phone(
    user_id=user['user_id'],
    phone_id=phone['phone_id'],
)
print('Added phone', phone['number'], 'to user', user['username'])

# Send two SMS messages to the phone with information about installing
# the app for PHONE_PLATFORM and activating it with this Duo account.
act_sent = admin_api.send_sms_activation_to_phone(
    phone_id=phone['phone_id'],
    install='1',
)
print('SMS activation sent to', phone['number'] + ':')
pprint.pprint(act_sent)



import unittest
from unittest.mock import patch
from .. import util
import duo_client.admin


class TestAccountAdmin(unittest.TestCase):

    def setUp(self):
        child_host = 'example2.com'
        kwargs = {'ikey': 'test_ikey', 'skey': 'test_skey', 'host': 'example.com'}

        patcher = patch("duo_client.admin.AccountAdmin.get_child_api_host")
        self.mock_child_host = patcher.start()
        self.mock_child_host.return_value = child_host
        self.addCleanup(patcher.stop)

        self.client = duo_client.admin.AccountAdmin(
            'DA012345678901234567', **kwargs)
        
        # monkeypatch client's _connect()
        self.client._connect = lambda: util.MockHTTPConnection()


if __name__ == '__main__':
    unittest.main()



import json
from .. import util
from .base import TestAdmin


class TestVerificationPush(TestAdmin):
    def test_send_verification_push(self):
        """
        Test sending a verification push to a user.
        """
        response = self.client.send_verification_push('test_user_id', 'test_phone_id')
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'],
                         '/admin/v1/users/test_user_id/send_verification_push')
        self.assertEqual(
           json.loads(response['body']),
            {'phone_id': 'test_phone_id', 'account_id': self.client.account_id})

    def test_get_verification_push_response(self):
        """
        Test getting the verification push response.
        """
        response = self.client.get_verification_push_response('test_user_id', 'test_push_id')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/users/test_user_id/verification_push_response')

        self.assertEqual(
            util.params_to_dict(args),
            {'push_id': ['test_push_id'], 'account_id': [self.client.account_id]})



import hashlib
from unittest import mock
import unittest
import duo_client.client
from . import util
import base64
import collections
import json

JSON_BODY = {
            'data': 'abc123',
            'alpha': ['a', 'b', 'c', 'd'],
            'info': {
                'test': 1,
                'another': 2,
            }
        }
JSON_STRING = '{"alpha":["a","b","c","d"],"data":"abc123","info":{"another":2,"test":1}}'


class TestQueryParameters(unittest.TestCase):
    """
    Tests for the proper canonicalization of query parameters for signing.
    """
    def assert_canon_params(self, params, expected):
        params = duo_client.client.normalize_params(params)
        self.assertEqual(
            duo_client.client.canon_params(params),
            expected,
        )

    def test_zero_params(self):
        self.assert_canon_params(
            {},
            '',
        )

    def test_one_param(self):
        self.assert_canon_params(
            {'realname': ['First Last']},
            'realname=First%20Last',
        )

    def test_two_params(self):
        self.assert_canon_params(
            {'realname': ['First Last'], 'username': ['root']},
            'realname=First%20Last&username=root')

    def test_with_boolean_true_int_and_string(self):
        self.assert_canon_params(
            {'words': ['First Last'], 'success': [True], 'digit': [5]},
            'digit=5&success=true&words=First%20Last')

    def test_with_boolean_false_int_and_string(self):
        self.assert_canon_params(
            {'words': ['First Last'], 'success': [False], 'digit': [5]},
            'digit=5&success=false&words=First%20Last')

    def test_list_string(self):
        """ A list and a string will both get converted. """
        self.assert_canon_params(
            {'realname': 'First Last', 'username': ['root']},
            'realname=First%20Last&username=root')

    def test_printable_ascii_characters(self):
        self.assert_canon_params(
            {
                'digits': ['0123456789'],
                'letters': ['abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'],
                'punctuation': ['!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'],
                'whitespace': ['\t\n\x0b\x0c\r '],
            },
            'digits=0123456789&letters=abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ&punctuation=%21%22%23%24%25%26%27%28%29%2A%2B%2C-.%2F%3A%3B%3C%3D%3E%3F%40%5B%5C%5D%5E_%60%7B%7C%7D~&whitespace=%09%0A%0B%0C%0D%20'
        )

    def test_unicode_fuzz_values(self):
        self.assert_canon_params(
            {
                u'bar': [u'\u2815\uaaa3\u37cf\u4bb7\u36e9\ucc05\u668e\u8162\uc2bd\ua1f1'],
                u'baz': [u'\u0df3\u84bd\u5669\u9985\ub8a4\uac3a\u7be7\u6f69\u934a\ub91c'],
                u'foo': [u'\ud4ce\ud6d6\u7938\u50c0\u8a20\u8f15\ufd0b\u8024\u5cb3\uc655'],
                u'qux': [u'\u8b97\uc846-\u828e\u831a\uccca\ua2d4\u8c3e\ub8b2\u99be'],
            },
            'bar=%E2%A0%95%EA%AA%A3%E3%9F%8F%E4%AE%B7%E3%9B%A9%EC%B0%85%E6%9A%8E%E8%85%A2%EC%8A%BD%EA%87%B1&baz=%E0%B7%B3%E8%92%BD%E5%99%A9%E9%A6%85%EB%A2%A4%EA%B0%BA%E7%AF%A7%E6%BD%A9%E9%8D%8A%EB%A4%9C&foo=%ED%93%8E%ED%9B%96%E7%A4%B8%E5%83%80%E8%A8%A0%E8%BC%95%EF%B4%8B%E8%80%A4%E5%B2%B3%EC%99%95&qux=%E8%AE%97%EC%A1%86-%E8%8A%8E%E8%8C%9A%EC%B3%8A%EA%8B%94%E8%B0%BE%EB%A2%B2%E9%A6%BE',
        )

    def test_unicode_fuzz_keys_and_values(self):
        self.assert_canon_params(
            {
                u'\u469a\u287b\u35d0\u8ef3\u6727\u502a\u0810\ud091\xc8\uc170': [u'\u0f45\u1a76\u341a\u654c\uc23f\u9b09\uabe2\u8343\u1b27\u60d0'],
                u'\u7449\u7e4b\uccfb\u59ff\ufe5f\u83b7\uadcc\u900c\ucfd1\u7813': [u'\u8db7\u5022\u92d3\u42ef\u207d\u8730\uacfe\u5617\u0946\u4e30'],
                u'\u7470\u9314\u901c\u9eae\u40d8\u4201\u82d8\u8c70\u1d31\ua042': [u'\u17d9\u0ba8\u9358\uaadf\ua42a\u48be\ufb96\u6fe9\ub7ff\u32f3'],
                u'\uc2c5\u2c1d\u2620\u3617\u96b3F\u8605\u20e8\uac21\u5934': [u'\ufba9\u41aa\ubd83\u840b\u2615\u3e6e\u652d\ua8b5\ud56bU'],
            },
            '%E4%9A%9A%E2%A1%BB%E3%97%90%E8%BB%B3%E6%9C%A7%E5%80%AA%E0%A0%90%ED%82%91%C3%88%EC%85%B0=%E0%BD%85%E1%A9%B6%E3%90%9A%E6%95%8C%EC%88%BF%E9%AC%89%EA%AF%A2%E8%8D%83%E1%AC%A7%E6%83%90&%E7%91%89%E7%B9%8B%EC%B3%BB%E5%A7%BF%EF%B9%9F%E8%8E%B7%EA%B7%8C%E9%80%8C%EC%BF%91%E7%A0%93=%E8%B6%B7%E5%80%A2%E9%8B%93%E4%8B%AF%E2%81%BD%E8%9C%B0%EA%B3%BE%E5%98%97%E0%A5%86%E4%B8%B0&%E7%91%B0%E9%8C%94%E9%80%9C%E9%BA%AE%E4%83%98%E4%88%81%E8%8B%98%E8%B1%B0%E1%B4%B1%EA%81%82=%E1%9F%99%E0%AE%A8%E9%8D%98%EA%AB%9F%EA%90%AA%E4%A2%BE%EF%AE%96%E6%BF%A9%EB%9F%BF%E3%8B%B3&%EC%8B%85%E2%B0%9D%E2%98%A0%E3%98%97%E9%9A%B3F%E8%98%85%E2%83%A8%EA%B0%A1%E5%A4%B4=%EF%AE%A9%E4%86%AA%EB%B6%83%E8%90%8B%E2%98%95%E3%B9%AE%E6%94%AD%EA%A2%B5%ED%95%ABU',
        )

    def test_sort_order_with_common_prefix(self):
        self.assert_canon_params(
            {
                'foo_bar': '2',
                'foo': '1',
            },
            'foo=1&foo_bar=2',
        )


class TestCanonicalize(unittest.TestCase):
    """
    Tests of the canonicalization of request attributes and parameters
    for signing.
    """
    def test_v1(self):
        test = {
            'host': 'foO.BAr52.cOm',
            'method': 'PoSt',
            'params': {
                u'\u469a\u287b\u35d0\u8ef3\u6727\u502a\u0810\ud091\xc8\uc170': [u'\u0f45\u1a76\u341a\u654c\uc23f\u9b09\uabe2\u8343\u1b27\u60d0'],
                u'\u7449\u7e4b\uccfb\u59ff\ufe5f\u83b7\uadcc\u900c\ucfd1\u7813': [u'\u8db7\u5022\u92d3\u42ef\u207d\u8730\uacfe\u5617\u0946\u4e30'],
                u'\u7470\u9314\u901c\u9eae\u40d8\u4201\u82d8\u8c70\u1d31\ua042': [u'\u17d9\u0ba8\u9358\uaadf\ua42a\u48be\ufb96\u6fe9\ub7ff\u32f3'],
                u'\uc2c5\u2c1d\u2620\u3617\u96b3F\u8605\u20e8\uac21\u5934': [u'\ufba9\u41aa\ubd83\u840b\u2615\u3e6e\u652d\ua8b5\ud56bU'],
            },
            'uri': '/Foo/BaR2/qux',
        }
        test['params'] = duo_client.client.normalize_params(test['params'])
        self.assertEqual(duo_client.client.canonicalize(sig_version=1,
                                                        date=None,
                                                        **test),
                         'POST\nfoo.bar52.com\n/Foo/BaR2/qux\n%E4%9A%9A%E2%A1%BB%E3%97%90%E8%BB%B3%E6%9C%A7%E5%80%AA%E0%A0%90%ED%82%91%C3%88%EC%85%B0=%E0%BD%85%E1%A9%B6%E3%90%9A%E6%95%8C%EC%88%BF%E9%AC%89%EA%AF%A2%E8%8D%83%E1%AC%A7%E6%83%90&%E7%91%89%E7%B9%8B%EC%B3%BB%E5%A7%BF%EF%B9%9F%E8%8E%B7%EA%B7%8C%E9%80%8C%EC%BF%91%E7%A0%93=%E8%B6%B7%E5%80%A2%E9%8B%93%E4%8B%AF%E2%81%BD%E8%9C%B0%EA%B3%BE%E5%98%97%E0%A5%86%E4%B8%B0&%E7%91%B0%E9%8C%94%E9%80%9C%E9%BA%AE%E4%83%98%E4%88%81%E8%8B%98%E8%B1%B0%E1%B4%B1%EA%81%82=%E1%9F%99%E0%AE%A8%E9%8D%98%EA%AB%9F%EA%90%AA%E4%A2%BE%EF%AE%96%E6%BF%A9%EB%9F%BF%E3%8B%B3&%EC%8B%85%E2%B0%9D%E2%98%A0%E3%98%97%E9%9A%B3F%E8%98%85%E2%83%A8%EA%B0%A1%E5%A4%B4=%EF%AE%A9%E4%86%AA%EB%B6%83%E8%90%8B%E2%98%95%E3%B9%AE%E6%94%AD%EA%A2%B5%ED%95%ABU')

    def test_v2(self):
        test = {
            'date': 'Fri, 07 Dec 2012 17:18:00 -0000',
            'host': 'foO.BAr52.cOm',
            'method': 'PoSt',
            'params': {u'\u469a\u287b\u35d0\u8ef3\u6727\u502a\u0810\ud091\xc8\uc170': [u'\u0f45\u1a76\u341a\u654c\uc23f\u9b09\uabe2\u8343\u1b27\u60d0'],
                       u'\u7449\u7e4b\uccfb\u59ff\ufe5f\u83b7\uadcc\u900c\ucfd1\u7813': [u'\u8db7\u5022\u92d3\u42ef\u207d\u8730\uacfe\u5617\u0946\u4e30'],
                       u'\u7470\u9314\u901c\u9eae\u40d8\u4201\u82d8\u8c70\u1d31\ua042': [u'\u17d9\u0ba8\u9358\uaadf\ua42a\u48be\ufb96\u6fe9\ub7ff\u32f3'],
                       u'\uc2c5\u2c1d\u2620\u3617\u96b3F\u8605\u20e8\uac21\u5934': [u'\ufba9\u41aa\ubd83\u840b\u2615\u3e6e\u652d\ua8b5\ud56bU']},
            'uri': '/Foo/BaR2/qux',
        }
        test['params'] = duo_client.client.normalize_params(test['params'])
        self.assertEqual(duo_client.client.canonicalize(sig_version=2,
                                                        **test),
                         'Fri, 07 Dec 2012 17:18:00 -0000\nPOST\nfoo.bar52.com\n/Foo/BaR2/qux\n%E4%9A%9A%E2%A1%BB%E3%97%90%E8%BB%B3%E6%9C%A7%E5%80%AA%E0%A0%90%ED%82%91%C3%88%EC%85%B0=%E0%BD%85%E1%A9%B6%E3%90%9A%E6%95%8C%EC%88%BF%E9%AC%89%EA%AF%A2%E8%8D%83%E1%AC%A7%E6%83%90&%E7%91%89%E7%B9%8B%EC%B3%BB%E5%A7%BF%EF%B9%9F%E8%8E%B7%EA%B7%8C%E9%80%8C%EC%BF%91%E7%A0%93=%E8%B6%B7%E5%80%A2%E9%8B%93%E4%8B%AF%E2%81%BD%E8%9C%B0%EA%B3%BE%E5%98%97%E0%A5%86%E4%B8%B0&%E7%91%B0%E9%8C%94%E9%80%9C%E9%BA%AE%E4%83%98%E4%88%81%E8%8B%98%E8%B1%B0%E1%B4%B1%EA%81%82=%E1%9F%99%E0%AE%A8%E9%8D%98%EA%AB%9F%EA%90%AA%E4%A2%BE%EF%AE%96%E6%BF%A9%EB%9F%BF%E3%8B%B3&%EC%8B%85%E2%B0%9D%E2%98%A0%E3%98%97%E9%9A%B3F%E8%98%85%E2%83%A8%EA%B0%A1%E5%A4%B4=%EF%AE%A9%E4%86%AA%EB%B6%83%E8%90%8B%E2%98%95%E3%B9%AE%E6%94%AD%EA%A2%B5%ED%95%ABU')

    def test_v4_with_json(self):
        hashed_body = hashlib.sha512(JSON_STRING.encode('utf-8')).hexdigest()
        expected = (
            'Tue, 04 Jul 2017 14:12:00\n'
            'POST\n'
            'foo.bar52.com\n'
            '/Foo/BaR2/qux\n\n' + hashed_body)
        params = {}
        actual = duo_client.client.canonicalize(
            'POST', 'foO.BaR52.cOm', '/Foo/BaR2/qux', params,
            'Tue, 04 Jul 2017 14:12:00', sig_version=4, body=JSON_STRING)

        self.assertEqual(actual, expected)

    def test_v5_with_json(self):
        hashed_body = hashlib.sha512(JSON_STRING.encode('utf-8')).hexdigest()
        headers = {"X-Duo-Header-1": "header_value_1"}
        expected = (
                'Tue, 17 Nov 2020 14:12:00\n'
                'POST\n'
                'foo.bar52.com\n'
                '/Foo/BaR2/qux\n\n' + hashed_body
                +'\n630b4bfe7e9abd03da2eee8f0a5d4e60a254ec880a839bcc2223bb5b9443e8ef24d58f0'
                 '254f1f5934bf8c017ebd0fd5b1acf86766bdbe74185e712a4092df3ed')
        params = {}
        body = duo_client.client.Client.canon_json(JSON_BODY)
        actual = duo_client.client.canonicalize(
            'POST', 'foO.BaR52.cOm', '/Foo/BaR2/qux', params, 'Tue, 17 Nov 2020 14:12:00',
            sig_version=5, body=body, additional_headers=headers)
        self.assertEqual(actual, expected)

    def test_invalid_signature_version_raises(self):
        params = duo_client.client.Client.canon_json(JSON_BODY)
        with self.assertRaises(ValueError) as e:
            duo_client.client.canonicalize(
                'POST', 'foO.BaR52.cOm', '/Foo/BaR2/qux', params,
                'Tue, 04 Jul 2017 14:12:00', sig_version=999)
        self.assertEqual(
            e.exception.args[0],
            "Unknown signature version: {}".format(999))

    def test_signature_v5_lowers_and_then_sorts_headers(self):
        hashed_body = hashlib.sha512(JSON_STRING.encode('utf-8')).hexdigest()
        headers = {
            "x-duo-A": "header_value_1",
            "X-Duo-B": "header_value_2"

        }
        expected = (
                'Tue, 17 Nov 2020 14:12:00\n'
                'POST\n'
                'foo.bar52.com\n'
                '/Foo/BaR2/qux\n\n'
                + hashed_body
                +'\n60be11a30e0756f2ee2afdce1db849b987dcf86c1133394b'
                'd7bbbc9877920330c4d78aceacbb377ab8cbd9a8efe6a410fed4047376635ac71226ab46ca10d2b1')
        params = {}
        body = duo_client.client.Client.canon_json(JSON_BODY)
        actual = duo_client.client.canonicalize(
            'POST', 'foO.BaR52.cOm', '/Foo/BaR2/qux', params, 'Tue, 17 Nov 2020 14:12:00',
            sig_version=5, body=body, additional_headers=headers)
        self.assertEqual(actual, expected)


class TestNormalizePageArgs(unittest.TestCase):

    def setUp(self):
        self.client = duo_client.client.Client(
            'test_ikey', 'test_akey', 'example.com')

    def test_normalize_page_args(self):

        tests = [
            (
                {},
                (None, '0')
            ),
            (
                {'offset': 9001},
                (None, '9001'),
            ),
            (
                {'limit': 2},
                ('2', '0'),
            ),
            (
                {'limit': '3'},
                ('3', '0'),
            ),
            (
                {'limit': 5, 'offset': 9002},
                ('5', '9002')
            )
        ]

        for (input, expected) in tests:
            output = self.client.normalize_paging_args(**input)
            self.assertEqual(output, expected)


class TestSign(unittest.TestCase):
    """
    Tests for proper signature creation for a request.
    """
    def test_hmac_sha1(self):
        test = {
            'date': 'Fri, 07 Dec 2012 17:18:00 -0000',
            'host': 'foO.BAr52.cOm',
            'method': 'PoSt',
            'params': {u'\u469a\u287b\u35d0\u8ef3\u6727\u502a\u0810\ud091\xc8\uc170': [u'\u0f45\u1a76\u341a\u654c\uc23f\u9b09\uabe2\u8343\u1b27\u60d0'],
                       u'\u7449\u7e4b\uccfb\u59ff\ufe5f\u83b7\uadcc\u900c\ucfd1\u7813': [u'\u8db7\u5022\u92d3\u42ef\u207d\u8730\uacfe\u5617\u0946\u4e30'],
                       u'\u7470\u9314\u901c\u9eae\u40d8\u4201\u82d8\u8c70\u1d31\ua042': [u'\u17d9\u0ba8\u9358\uaadf\ua42a\u48be\ufb96\u6fe9\ub7ff\u32f3'],
                       u'\uc2c5\u2c1d\u2620\u3617\u96b3F\u8605\u20e8\uac21\u5934': [u'\ufba9\u41aa\ubd83\u840b\u2615\u3e6e\u652d\ua8b5\ud56bU']},
            'uri': '/Foo/BaR2/qux',
        }
        test['params'] = duo_client.client.normalize_params(test['params'])
        ikey = 'test_ikey'
        actual = duo_client.client.sign(
            sig_version=2,
            ikey=ikey,
            skey='gtdfxv9YgVBYcF6dl2Eq17KUQJN2PLM2ODVTkvoT',
            digestmod=hashlib.sha1,
            **test
        )
        expected = 'f01811cbbf9561623ab45b893096267fd46a5178'
        expected = ikey + ':' + expected
        if isinstance(expected, str):
            expected = expected.encode('utf-8')
        expected = base64.b64encode(expected).strip()
        if not isinstance(expected, str):
            expected = expected.decode('utf-8')
        expected = 'Basic ' + expected
        self.assertEqual(actual,
                         expected)

    def test_hmac_sha512(self):
        test = {
            'date': 'Fri, 07 Dec 2012 17:18:00 -0000',
            'host': 'foO.BAr52.cOm',
            'method': 'PoSt',
            'params': {u'\u469a\u287b\u35d0\u8ef3\u6727\u502a\u0810\ud091\xc8\uc170': [u'\u0f45\u1a76\u341a\u654c\uc23f\u9b09\uabe2\u8343\u1b27\u60d0'],
                       u'\u7449\u7e4b\uccfb\u59ff\ufe5f\u83b7\uadcc\u900c\ucfd1\u7813': [u'\u8db7\u5022\u92d3\u42ef\u207d\u8730\uacfe\u5617\u0946\u4e30'],
                       u'\u7470\u9314\u901c\u9eae\u40d8\u4201\u82d8\u8c70\u1d31\ua042': [u'\u17d9\u0ba8\u9358\uaadf\ua42a\u48be\ufb96\u6fe9\ub7ff\u32f3'],
                       u'\uc2c5\u2c1d\u2620\u3617\u96b3F\u8605\u20e8\uac21\u5934': [u'\ufba9\u41aa\ubd83\u840b\u2615\u3e6e\u652d\ua8b5\ud56bU']},
            'uri': '/Foo/BaR2/qux',
        }
        test['params'] = duo_client.client.normalize_params(test['params'])
        ikey = 'test_ikey'
        actual = duo_client.client.sign(
            sig_version=2,
            ikey=ikey,
            skey='gtdfxv9YgVBYcF6dl2Eq17KUQJN2PLM2ODVTkvoT',
            **test
        )
        expected = '0508065035a03b2a1de2f453e629e791d180329e157f65df6b3e0f08299d4321e1c5c7a7c7ee6b9e5fc80d1fb6fbf3ad5eb7c44dd3b3985a02c37aca53ec3698'
        expected = ikey + ':' + expected
        if isinstance(expected, str):
            expected = expected.encode('utf-8')
        expected = base64.b64encode(expected).strip()
        if not isinstance(expected, str):
            expected = expected.decode('utf-8')
        expected = 'Basic ' + expected
        self.assertEqual(actual,
                         expected)

class TestRequest(unittest.TestCase):
    """ Tests for the request created by api_call and json_api_call. """
    # usful args for testing
    args_in = {
        'foo':['bar'],
        'baz':['qux', 'quux=quuux', 'foobar=foobar&barbaz=barbaz']}
    args_out = dict(
        (key, [v for v in val])
        for (key, val) in list(args_in.items()))

    def setUp(self):
        self.client = duo_client.client.Client(
            'test_ikey', 'test_akey', 'example.com')
        # monkeypatch client's _connect()
        self.client._connect = lambda: util.MockHTTPConnection()

    def test_api_call_get_no_params(self):
        (response, dummy) = self.client.api_call('GET', '/foo/bar', {})
        self.assertEqual(response.method, 'GET')
        self.assertEqual(response.uri, '/foo/bar?')

    def test_api_call_post_no_params(self):
        (response, dummy) = self.client.api_call('POST', '/foo/bar', {})
        self.assertEqual(response.method, 'POST')
        self.assertEqual(response.uri, '/foo/bar')
        self.assertEqual(json.loads(response.body), {})

    def test_api_call_get_params(self):
        (response, dummy) = self.client.api_call(
            'GET', '/foo/bar', self.args_in)
        self.assertEqual(response.method, 'GET')
        (uri, args) = response.uri.split('?')
        self.assertEqual(uri, '/foo/bar')
        self.assertEqual(util.params_to_dict(args), self.args_out)

    def test_api_call_post_params(self):
        (response, dummy) = self.client.api_call(
            'POST', '/foo/bar', self.args_in)
        self.assertEqual(response.method, 'POST')
        self.assertEqual(response.uri, '/foo/bar')
        self.assertEqual(json.loads(response.body), self.args_out)

    def test_json_api_call_get_no_params(self):
        response = self.client.json_api_call('GET', '/foo/bar', {})
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(response['uri'], '/foo/bar?')
        self.assertEqual(response['body'], None)

    def test_json_api_call_post_no_params(self):
        response = self.client.json_api_call('POST', '/foo/bar', {})
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/foo/bar')
        self.assertEqual(json.loads(response['body']), {})

    def test_json_api_call_get_params(self):
        response = self.client.json_api_call(
            'GET', '/foo/bar', self.args_in)
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/foo/bar')
        self.assertEqual(util.params_to_dict(args), self.args_out)

    def test_json_api_call_post_params(self):
        response = self.client.json_api_call(
            'POST', '/foo/bar', self.args_in)
        self.assertEqual(response['method'], 'POST')
        self.assertEqual(response['uri'], '/foo/bar')
        self.assertEqual(json.loads(response['body']), self.args_out)

class TestPaging(unittest.TestCase):
    def setUp(self):
        self.client = util.CountingClient(
            'test_ikey', 'test_akey', 'example.com', paging_limit=100)
        self.objects = [util.MockJsonObject() for i in range(1000)]
        self.client._connect = lambda: util.MockPagingHTTPConnection(self.objects)

    def test_get_objects_paging(self):
        response = self.client.json_paging_api_call(
            'GET', '/admin/v1/objects', {})
        self.assertEqual(len(self.objects), len(list(response)))
        self.assertEqual(10, self.client.counter)

    def test_get_no_objects_paging(self):
        self.objects = []
        self.client._connect = lambda: util.MockPagingHTTPConnection(self.objects)
        response = self.client.json_paging_api_call(
            'GET', '/admin/v1/objects', {})
        self.assertEqual(len(self.objects), len(list(response)))
        self.assertEqual(1, self.client.counter)

    def test_get_objects_paging_limit(self):
        response = self.client.json_paging_api_call(
            'GET', '/admin/v1/objects', {'limit':'250'})
        self.assertEqual(len(self.objects), len(list(response)))
        self.assertEqual(4, self.client.counter)

    def test_get_all_objects(self):
        response = self.client.json_paging_api_call(
            'GET', '/admin/v1/objects', {'limit':'1000'})
        expected = [obj.to_json() for obj in self.objects]
        self.assertListEqual(expected, list(response))
        self.assertEqual(1, self.client.counter)

class TestAlternatePaging(unittest.TestCase):
    def setUp(self):
        self.client = util.CountingClient(
            'test_ikey', 'test_akey', 'example.com', paging_limit=100)
        self.objects = [util.MockJsonObject() for i in range(1000)]
        self.client._connect = lambda: util.MockAlternatePagingHTTPConnection(self.objects)

    def test_get_objects_paging(self):
        response = self.client.json_cursor_api_call(
            'GET', '/admin/v1/objects', {},
            lambda response: response['data']
        )
        self.assertEqual(len(self.objects), len(list(response)))
        self.assertEqual(10, self.client.counter)

    def test_get_no_objects_paging(self):
        self.objects = []
        self.client._connect = lambda: util.MockAlternatePagingHTTPConnection(self.objects)
        response = self.client.json_cursor_api_call(
            'GET', '/admin/v1/objects', {},
            lambda response: response['data']
        )
        self.assertEqual(len(self.objects), len(list(response)))
        self.assertEqual(1, self.client.counter)

    def test_get_objects_paging_limit(self):
        response = self.client.json_cursor_api_call(
            'GET', '/admin/v1/objects', {'limit':'250'},
            lambda response: response['data']
        )
        self.assertEqual(len(self.objects), len(list(response)))
        self.assertEqual(4, self.client.counter)

    def test_get_all_objects(self):
        response = self.client.json_cursor_api_call(
            'GET', '/admin/v1/objects', {'limit':'1000'},
            lambda response: response['data']
        )
        expected = [obj.to_json() for obj in self.objects]
        self.assertListEqual(expected, list(response))
        self.assertEqual(1, self.client.counter)

class TestRequestsV4(unittest.TestCase):
    # usful args for testing GETs
    args_in = {
        'foo':['bar'],
        'baz':['qux', 'quux=quuux', 'foobar=foobar&barbaz=barbaz']}
    args_out = dict(
        (key, [v for v in val])
        for (key, val) in list(args_in.items()))

    def setUp(self):
        self.client = duo_client.client.Client(
            'test_ikey', 'test_akey', 'example.com', sig_timezone='America/Detroit',
            digestmod=hashlib.sha512, sig_version=4)
        # monkeypatch client's _connect()
        self.client._connect = lambda: util.MockHTTPConnection()

    def test_get_no_params(self):
        (response, dummy) = self.client.api_call('GET', '/foo/bar', {})
        self.assertEqual(response.method, 'GET')
        self.assertEqual(response.uri, '/foo/bar?')
        self.assertIn('Authorization', response.headers)

    def test_get_params(self):
        (response, dummy) = self.client.api_call(
            'GET', '/foo/bar', self.args_in)
        self.assertEqual(response.method, 'GET')
        (uri, args) = response.uri.split('?')
        self.assertEqual(uri, '/foo/bar')
        self.assertEqual(util.params_to_dict(args), self.args_out)
        self.assertIn('Authorization', response.headers)

    def test_json_api_call_get_no_params(self):
        response = self.client.json_api_call('GET', '/foo/bar', {})
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(response['uri'], '/foo/bar?')
        self.assertEqual(response['body'], None)
        self.assertIn('Authorization', response['headers'])

    def test_json_api_call_get_params(self):
        response = self.client.json_api_call(
            'GET', '/foo/bar', self.args_in)
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/foo/bar')
        self.assertEqual(util.params_to_dict(args), self.args_out)
        self.assertIn('Authorization', response['headers'])

    def test_json_post(self):
        (response, dummy) = self.client.api_call('POST', '/foo/bar', JSON_BODY)

        self.assertEqual(response.method, 'POST')
        self.assertEqual(response.uri, '/foo/bar')
        self.assertEqual(response.body, JSON_STRING)

        self.assertIn('Content-type', response.headers)
        self.assertEqual(response.headers['Content-type'], 'application/json')
        self.assertIn('Authorization', response.headers)

    def test_json_fails_with_bad_args(self):
        with self.assertRaises(ValueError) as e:
            (response, dummy) = self.client.api_call('POST', '/foo/bar', '')
        self.assertEqual(e.exception.args[0], "JSON request must be an object.")

    def test_json_put(self):
        (response, dummy) = self.client.api_call('PUT', '/foo/bar', JSON_BODY)

        self.assertEqual(response.method, 'PUT')
        self.assertEqual(response.uri, '/foo/bar')
        self.assertEqual(response.body, JSON_STRING)

        self.assertIn('Content-type', response.headers)
        self.assertEqual(response.headers['Content-type'], 'application/json')
        self.assertIn('Authorization', response.headers)

class TestParseJsonResponse(unittest.TestCase):
    APIResponse = collections.namedtuple('APIResponse', 'status reason')

    def setUp(self):
        self.client = duo_client.client.Client(
            'test_ikey', 'test_akey', 'example.com', sig_timezone='America/Detroit',
            sig_version=2)

    def test_good_response(self):
        api_res = self.APIResponse(200, '')
        expected_data = {
            'foo': 'bar'
        }
        res_body = {
            'response': expected_data,
            'stat': 'OK'
        }

        data = self.client.parse_json_response(api_res, json.dumps(res_body))

        self.assertEqual(data, expected_data)

    def test_response_contains_invalid_json(self):
        api_res = self.APIResponse(200, 'Fake reason')

        response = 'Bad JSON'
        with self.assertRaises(RuntimeError) as e:
            self.client.parse_json_response(api_res, response)

        self.assertEqual(e.exception.status, api_res.status)
        self.assertEqual(e.exception.reason, api_res.reason)
        self.assertEqual(e.exception.data, response)

    def test_response_stat_isnot_OK(self):
        api_res = self.APIResponse(200, 'Fake reason')

        res_body = {
                'response': {
                'foo': 'bar'
            },
            'stat': 'FAIL'
        }

        with self.assertRaises(RuntimeError) as e:
            self.client.parse_json_response(api_res, json.dumps(res_body))

        self.assertEqual(e.exception.status, api_res.status)
        self.assertEqual(e.exception.reason, api_res.reason)
        self.assertEqual(e.exception.data, res_body)

    def test_response_is_http_error(self):
        for code in range(201, 600):
            api_res = self.APIResponse(code, 'fake reason')
            res_body = {
                'response': 'some message',
                'stat': 'OK'
            }
            with self.assertRaises(RuntimeError) as e:
                self.client.parse_json_response(api_res, json.dumps(res_body))

            self.assertEqual(e.exception.status, api_res.status)
            self.assertEqual(e.exception.reason, api_res.reason)
            self.assertEqual(e.exception.data, res_body)

class TestParseJsonResponseAndMetadata(unittest.TestCase):
    APIResponse = collections.namedtuple('APIResponse', 'status reason')

    def setUp(self):
        self.client = duo_client.client.Client(
            'test_ikey', 'test_akey', 'example.com', sig_timezone='America/Detroit',
            sig_version=2)

    def test_good_response(self):
        api_res = self.APIResponse(200, '')
        expected_data = {
            'foo': 'bar'
        }
        res_body = {
            'response': expected_data,
            'stat': 'OK'
        }

        data, metadata = self.client.parse_json_response_and_metadata(api_res, json.dumps(res_body))

        self.assertEqual(data, expected_data)
        self.assertEqual(metadata, {})

    def test_response_contains_invalid_json(self):
        api_res = self.APIResponse(200, 'Fake reason')

        response = 'Bad JSON'
        with self.assertRaises(RuntimeError) as e:
            self.client.parse_json_response_and_metadata(api_res, response)

        self.assertEqual(e.exception.status, api_res.status)
        self.assertEqual(e.exception.reason, api_res.reason)
        self.assertEqual(e.exception.data, response)

    def test_response_is_a_string(self):
        """
        Some API requests return just a string in their response.
        We want to make sure we handle those correctly
        """
        api_res = self.APIResponse(200, 'Fake reason')
        response = {
            'response': "just a string",
            'stat': 'OK'
        }
        try:
            self.client.parse_json_response_and_metadata(api_res, json.dumps(response))
        except Exception:
            self.fail("parsing raised exception for string response")

    def test_response_is_a_list(self):
        """
        Some API requests return just a list in their response. with
        metadata included at the top level.
        We want to make sure we handle those correctly
        """
        api_res = self.APIResponse(200, "Fake reason")
        expected_metadata = { "offset": 4 }
        expected_response = ["multiple", "elements"]
        response = {
            "response": expected_response,
            "metadata": expected_metadata,
            "stat": "OK"
        }
        response, metadata = self.client.parse_json_response_and_metadata(api_res, json.dumps(response))
        self.assertEqual(metadata, expected_metadata)

    def test_response_stat_isnot_OK(self):
        api_res = self.APIResponse(200, 'Fake reason')

        res_body = {
                'response': {
                'foo': 'bar'
            },
            'stat': 'FAIL'
        }

        with self.assertRaises(RuntimeError) as e:
            self.client.parse_json_response_and_metadata(api_res, json.dumps(res_body))

        self.assertEqual(e.exception.status, api_res.status)
        self.assertEqual(e.exception.reason, api_res.reason)
        self.assertEqual(e.exception.data, res_body)

    def test_response_is_http_error(self):
        for code in range(201, 600):
            api_res = self.APIResponse(code, 'fake reason')
            res_body = {
                'response': 'some message',
                'stat': 'OK'
            }
            with self.assertRaises(RuntimeError) as e:
                self.client.parse_json_response_and_metadata(api_res, json.dumps(res_body))

            self.assertEqual(e.exception.status, api_res.status)
            self.assertEqual(e.exception.reason, api_res.reason)
            self.assertEqual(e.exception.data, res_body)

@mock.patch('duo_client.client.sleep')
class TestRetryRequests(unittest.TestCase):
    def setUp(self):
        self.client = duo_client.client.Client(
            'test_ikey', 'test_akey', 'example.com',
        )

    def test_non_limited_reponse(self, mock_sleep):
        # monkeypatch client's _connect()
        mock_connection = util.MockMultipleRequestHTTPConnection(
            [200])
        self.client._connect = lambda: mock_connection
        (response, dummy) = self.client.api_call('GET', '/foo/bar', {})
        mock_sleep.assert_not_called()

        self.assertEqual(response.status, 200)
        self.assertEqual(mock_connection.requests, 1)

    @mock.patch('duo_client.client.random')
    def test_single_limited_response(self, mock_random, mock_sleep):
        mock_random.uniform.return_value = 0.123
        # monkeypatch client's _connect()
        mock_connection = util.MockMultipleRequestHTTPConnection(
            [429, 200])
        self.client._connect = lambda: mock_connection

        (response, dummy) = self.client.api_call('GET', '/foo/bar', {})

        mock_sleep.assert_called_once_with(1.123)
        mock_random.uniform.assert_called_once()
        self.assertEqual(response.status, 200)
        self.assertEqual(mock_connection.requests, 2)

    @mock.patch('duo_client.client.random')
    def test_all_limited_responses(self, mock_random, mock_sleep):
        mock_random.uniform.return_value = 0.123
        # monkeypatch client's _connect()
        mock_connection = util.MockMultipleRequestHTTPConnection(
            [429, 429, 429, 429, 429, 429, 429])
        self.client._connect = lambda: mock_connection

        (response, data) = self.client.api_call('GET', '/foo/bar', {})

        expected_sleep_calls = (
            mock.call(1.123),
            mock.call(2.123),
            mock.call(4.123),
            mock.call(8.123),
            mock.call(16.123),
            mock.call(32.123),
        )
        mock_sleep.assert_has_calls(expected_sleep_calls)
        expected_random_calls = (
            mock.call(0, 1),
            mock.call(0, 1),
            mock.call(0, 1),
            mock.call(0, 1),
            mock.call(0, 1),
            mock.call(0, 1),
        )
        mock_random.uniform.assert_has_calls(expected_random_calls)
        self.assertEqual(response.status, 429)
        self.assertEqual(mock_connection.requests, 7)

class TestInstantiate(unittest.TestCase):
    def test_sig_version_3_raises_exception(self):
        with self.assertRaises(ValueError):
            duo_client.client.Client(
                'test_ikey', 'test_akey', 'example.com', sig_timezone='America/Detroit',
                sig_version=3)

if __name__ == '__main__':
    unittest.main()



import unittest

from .. import util
import duo_client.admin
from .base import TestAdmin


class TestEndpoints(TestAdmin):
    def test_get_endpoints_iterator(self):
        """ Test to get endpoints iterator
        """
        iterator = self.client_list.get_endpoints_iterator()
        response = next(iterator)
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/endpoints')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_endpoints(self):
        """ Test to get endpoints.
        """
        response = self.client_list.get_endpoints()[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/endpoints')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_endpoints_offset(self):
        """ Test to get endpoints with pagination params.
        """
        response = self.client_list.get_endpoints(offset=20)[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/endpoints')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_endpoints_limit(self):
        """ Test to get endpoints with pagination params.
        """
        response = self.client_list.get_endpoints(limit=20)[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/endpoints')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['20'],
                'offset': ['0'],
            })

    def test_get_endpoints_limit_and_offset(self):
        """ Test to get endpoints with pagination params.
        """
        response = self.client_list.get_endpoints(limit=35, offset=20)[0]
        self.assertEqual(response['method'], 'GET')
        (uri, args) = response['uri'].split('?')
        self.assertEqual(uri, '/admin/v1/endpoints')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
                'limit': ['35'],
                'offset': ['20'],
            })

    def test_get_endpoint(self):
        """ Test getting a single endpoint.
        """
        epkey = 'EP18JX1A10AB102M2T2X'
        response = self.client_list.get_endpoint(epkey)[0]
        (uri, args) = response['uri'].split('?')
        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/endpoints/' + epkey)
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client.account_id],
            })

if __name__ == '__main__':
    unittest.main()



"""
Example of Duo Accounts API child account deletiom
"""

import duo_client
import sys
import getpass

from pprint import pprint


argv_iter = iter(sys.argv[1:])


def _get_next_arg(prompt, secure=False):
    """Read information from STDIN, using getpass when sensitive information should not be echoed to tty"""
    try:
        return next(argv_iter)
    except StopIteration:
        if secure is True:
            return getpass.getpass(prompt)
        else:
            return input(prompt)


def prompt_for_credentials() -> dict:
    """Collect required API credentials from command line prompts

    :return: dictionary containing Duo Accounts API ikey, skey and hostname strings
    """

    ikey = _get_next_arg('Duo Accounts API integration key ("DI..."): ')
    skey = _get_next_arg('Duo Accounts API integration secret key: ', secure=True)
    host = _get_next_arg('Duo Accounts API hostname ("api-....duosecurity.com"): ')
    account_id = _get_next_arg('ID of child account to delete: ')

    return {"IKEY": ikey, "SKEY": skey, "APIHOST": host, "ACCOUNT_ID": account_id}


def main():
    """Main program entry point"""

    inputs = prompt_for_credentials()

    account_client = duo_client.Accounts(
            ikey=inputs['IKEY'],
            skey=inputs['SKEY'],
            host=inputs['APIHOST']
    )

    account_name = None
    child_account_list = account_client.get_child_accounts()
    for account in child_account_list:
        if account['account_id'] == inputs['ACCOUNT_ID']:
            account_name = account['name']
    if account_name is None:
        print(f"Unable to find account with ID [{inputs['ACCOUNT_ID']}]")
        sys.exit()

    print(f"Deleting child account with name [{account_name}]")
    deleted_account = account_client.delete_account(inputs['ACCOUNT_ID'])
    if deleted_account == '':
        print(f"Account {inputs['ACCOUNT_ID']} was deleted successfully.")
    else:
        print(f"An unexpected error occurred while deleting account [{account_name}: {deleted_account}]")


if __name__ == '__main__':
    main()



from .. import util
import duo_client.admin
from .base import TestAdmin


class TestDesktopTokens(TestAdmin):
    def test_get_desktoptokens_generator(self):
        """ Test to get desktop tokens generator.
        """
        generator = self.client_list.get_desktoptokens_generator()
        response = next(generator)
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/desktoptokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_desktoptokens(self):
        """ Test to get desktop tokens without params.
        """
        response = self.client_list.get_desktoptokens()[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/desktoptokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_desktoptokens_limit(self):
        """ Test to get desktop tokens with limit.
        """
        response = self.client_list.get_desktoptokens(limit='20')[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/desktoptokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['20'],
                'offset': ['0'],
            })

    def test_get_desktoptokens_offset(self):
        """ Test to get desktop tokens with offset.
        """
        response = self.client_list.get_desktoptokens(offset='20')[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/desktoptokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['100'],
                'offset': ['0'],
            })

    def test_get_desktoptokens_limit_offset(self):
        """ Test to get desktop tokens with limit and offset.
        """
        response = self.client_list.get_desktoptokens(limit='20', offset='2')[0]
        uri, args = response['uri'].split('?')

        self.assertEqual(response['method'], 'GET')
        self.assertEqual(uri, '/admin/v1/desktoptokens')
        self.assertEqual(
            util.params_to_dict(args),
            {
                'account_id': [self.client_list.account_id],
                'limit': ['20'],
                'offset': ['2'],
            })
